# -*- coding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from oslo.config import cfg
from oslo import messaging
import six
from stevedore import driver

from mistral import context as auth_context
from mistral import engine
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)


def get_executor(name, transport):
    mgr = driver.DriverManager(
        namespace='mistral.executor.drivers',
        name=name,
        invoke_on_load=True,
        invoke_kwds={'transport': transport})
    return mgr.driver


@six.add_metaclass(abc.ABCMeta)
class Executor(object):
    """Abstract class for task execution."""

    def __init__(self, transport=None):
        self.transport = engine.get_transport(transport)
        self.engine = engine.EngineClient(self.transport)

    @abc.abstractmethod
    def handle_task(self, cntx, **kwargs):
        raise NotImplementedError()


class ExecutorClient(object):
    """RPC client for the Executor."""

    def __init__(self, transport):
        """Construct an RPC client for the Executor.

        :param transport: a messaging transport handle
        :type transport: Transport
        """
        serializer = auth_context.RpcContextSerializer(
            auth_context.JsonPayloadSerializer())
        target = messaging.Target(topic=cfg.CONF.executor.topic)
        self._client = messaging.RPCClient(transport, target,
                                           serializer=serializer)

    def handle_task(self, cntx, **kwargs):
        """Send the task request to the Executor for execution.

        :param cntx: a request context dict
        :type cntx: MistralContext
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        """
        return self._client.cast(cntx, 'handle_task', **kwargs)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import eventlet
import mock
from oslo.config import cfg

eventlet.monkey_patch()

from mistral.actions import std_actions
from mistral.cmd import launch
from mistral.db import api as db_api
from mistral.engine import states
from mistral.openstack.common import importutils
from mistral.openstack.common import log as logging
from mistral.tests import base


# We need to make sure that all configuration properties are registered.
importutils.import_module("mistral.config")
LOG = logging.getLogger(__name__)

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

WB_NAME = 'my_workbook'
CONTEXT = None  # TODO(rakhmerov): Use a meaningful value.


class TestTransport(base.EngineTestCase):
    def setUp(self):
        super(TestTransport, self).setUp()

        # Run the Engine and Executor in the background.
        self.en_thread = eventlet.spawn(launch.launch_engine, self.transport)
        self.addCleanup(self.en_thread.kill)
        self.ex_thread = eventlet.spawn(launch.launch_executor, self.transport)
        self.addCleanup(self.ex_thread.kill)

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(
            return_value={'definition': base.get_resource('test_rest.yaml')}))
    @mock.patch.object(
        std_actions.HTTPAction, 'run', mock.MagicMock(return_value={}))
    def test_transport(self):
        """Test if engine request traversed through the oslo.messaging
        transport.
        """
        execution = self.engine.start_workflow_execution(
            WB_NAME, 'create-vms', CONTEXT)

        task = db_api.tasks_get(workbook_name=WB_NAME,
                                execution_id=execution['id'])[0]

        # Check task execution state. There is no timeout mechanism in
        # unittest. There is an example to add a custom timeout decorator that
        # can wrap this test function in another process and then manage the
        # process time. However, it seems more straightforward to keep the
        # loop finite.
        for i in range(0, 50):
            db_task = db_api.task_get(task['id'])
            # Ensure the request reached the executor and the action has ran.
            if db_task['state'] != states.IDLE:
                # We have to wait sometime due to time interval between set
                # task state to RUNNING and invocation action.run()
                time.sleep(0.1)
                self.assertIn(db_task['state'],
                              [states.RUNNING, states.SUCCESS, states.ERROR])
                return
            time.sleep(0.1)

        # Task is not being processed. Throw an exception here.
        raise Exception('Timed out waiting for task to be processed.')



# Copyright 2012 SINA Corporation
# Copyright 2014 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Extracts OpenStack config option info from module(s)."""

from __future__ import print_function

import argparse
import imp
import os
import re
import socket
import sys
import textwrap

from oslo.config import cfg
import six
import stevedore.named

from mistral.openstack.common import gettextutils
from mistral.openstack.common import importutils

gettextutils.install('mistral')

STROPT = "StrOpt"
BOOLOPT = "BoolOpt"
INTOPT = "IntOpt"
FLOATOPT = "FloatOpt"
LISTOPT = "ListOpt"
DICTOPT = "DictOpt"
MULTISTROPT = "MultiStrOpt"

OPT_TYPES = {
    STROPT: 'string value',
    BOOLOPT: 'boolean value',
    INTOPT: 'integer value',
    FLOATOPT: 'floating point value',
    LISTOPT: 'list value',
    DICTOPT: 'dict value',
    MULTISTROPT: 'multi valued',
}

OPTION_REGEX = re.compile(r"(%s)" % "|".join([STROPT, BOOLOPT, INTOPT,
                                              FLOATOPT, LISTOPT, DICTOPT,
                                              MULTISTROPT]))

PY_EXT = ".py"
BASEDIR = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       "../../../../"))
WORDWRAP_WIDTH = 60


def raise_extension_exception(extmanager, ep, err):
    raise


def generate(argv):
    parser = argparse.ArgumentParser(
        description='generate sample configuration file',
    )
    parser.add_argument('-m', dest='modules', action='append')
    parser.add_argument('-l', dest='libraries', action='append')
    parser.add_argument('srcfiles', nargs='*')
    parsed_args = parser.parse_args(argv)

    mods_by_pkg = dict()
    for filepath in parsed_args.srcfiles:
        pkg_name = filepath.split(os.sep)[1]
        mod_str = '.'.join(['.'.join(filepath.split(os.sep)[:-1]),
                            os.path.basename(filepath).split('.')[0]])
        mods_by_pkg.setdefault(pkg_name, list()).append(mod_str)
    # NOTE(lzyeval): place top level modules before packages
    pkg_names = sorted(pkg for pkg in mods_by_pkg if pkg.endswith(PY_EXT))
    ext_names = sorted(pkg for pkg in mods_by_pkg if pkg not in pkg_names)
    pkg_names.extend(ext_names)

    # opts_by_group is a mapping of group name to an options list
    # The options list is a list of (module, options) tuples
    opts_by_group = {'DEFAULT': []}

    if parsed_args.modules:
        for module_name in parsed_args.modules:
            module = _import_module(module_name)
            if module:
                for group, opts in _list_opts(module):
                    opts_by_group.setdefault(group, []).append((module_name,
                                                                opts))

    # Look for entry points defined in libraries (or applications) for
    # option discovery, and include their return values in the output.
    #
    # Each entry point should be a function returning an iterable
    # of pairs with the group name (or None for the default group)
    # and the list of Opt instances for that group.
    if parsed_args.libraries:
        loader = stevedore.named.NamedExtensionManager(
            'oslo.config.opts',
            names=list(set(parsed_args.libraries)),
            invoke_on_load=False,
            on_load_failure_callback=raise_extension_exception
        )
        for ext in loader:
            for group, opts in ext.plugin():
                opt_list = opts_by_group.setdefault(group or 'DEFAULT', [])
                opt_list.append((ext.name, opts))

    for pkg_name in pkg_names:
        mods = mods_by_pkg.get(pkg_name)
        mods.sort()
        for mod_str in mods:
            if mod_str.endswith('.__init__'):
                mod_str = mod_str[:mod_str.rfind(".")]

            mod_obj = _import_module(mod_str)
            if not mod_obj:
                raise RuntimeError("Unable to import module %s" % mod_str)

            for group, opts in _list_opts(mod_obj):
                opts_by_group.setdefault(group, []).append((mod_str, opts))

    print_group_opts('DEFAULT', opts_by_group.pop('DEFAULT', []))
    for group in sorted(opts_by_group.keys()):
        print_group_opts(group, opts_by_group[group])


def _import_module(mod_str):
    try:
        if mod_str.startswith('bin.'):
            imp.load_source(mod_str[4:], os.path.join('bin', mod_str[4:]))
            return sys.modules[mod_str[4:]]
        else:
            return importutils.import_module(mod_str)
    except Exception as e:
        sys.stderr.write("Error importing module %s: %s\n" % (mod_str, str(e)))
        return None


def _is_in_group(opt, group):
    "Check if opt is in group."
    for value in group._opts.values():
        # NOTE(llu): Temporary workaround for bug #1262148, wait until
        # newly released oslo.config support '==' operator.
        if not(value['opt'] != opt):
            return True
    return False


def _guess_groups(opt, mod_obj):
    # is it in the DEFAULT group?
    if _is_in_group(opt, cfg.CONF):
        return 'DEFAULT'

    # what other groups is it in?
    for value in cfg.CONF.values():
        if isinstance(value, cfg.CONF.GroupAttr):
            if _is_in_group(opt, value._group):
                return value._group.name

    raise RuntimeError(
        "Unable to find group for option %s, "
        "maybe it's defined twice in the same group?"
        % opt.name
    )


def _list_opts(obj):
    def is_opt(o):
        return (isinstance(o, cfg.Opt) and
                not isinstance(o, cfg.SubCommandOpt))

    opts = list()
    for attr_str in dir(obj):
        attr_obj = getattr(obj, attr_str)
        if is_opt(attr_obj):
            opts.append(attr_obj)
        elif (isinstance(attr_obj, list) and
              all(map(lambda x: is_opt(x), attr_obj))):
            opts.extend(attr_obj)

    ret = {}
    for opt in opts:
        ret.setdefault(_guess_groups(opt, obj), []).append(opt)
    return ret.items()


def print_group_opts(group, opts_by_module):
    print("[%s]" % group)
    print('')
    for mod, opts in opts_by_module:
        print('#')
        print('# Options defined in %s' % mod)
        print('#')
        print('')
        for opt in opts:
            _print_opt(opt)
        print('')


def _get_my_ip():
    try:
        csock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        csock.connect(('8.8.8.8', 80))
        (addr, port) = csock.getsockname()
        csock.close()
        return addr
    except socket.error:
        return None


def _sanitize_default(name, value):
    """Set up a reasonably sensible default for pybasedir, my_ip and host."""
    hostname = socket.gethostname()
    fqdn = socket.getfqdn()
    if value.startswith(sys.prefix):
        # NOTE(jd) Don't use os.path.join, because it is likely to think the
        # second part is an absolute pathname and therefore drop the first
        # part.
        value = os.path.normpath("/usr/" + value[len(sys.prefix):])
    elif value.startswith(BASEDIR):
        return value.replace(BASEDIR, '/usr/lib/python/site-packages')
    elif BASEDIR in value:
        return value.replace(BASEDIR, '')
    elif value == _get_my_ip():
        return '10.0.0.1'
    elif value in (hostname, fqdn):
        if 'host' in name:
            return 'mistral'
    elif value.endswith(hostname):
        return value.replace(hostname, 'mistral')
    elif value.endswith(fqdn):
        return value.replace(fqdn, 'mistral')
    elif value.strip() != value:
        return '"%s"' % value
    return value


def _print_opt(opt):
    opt_name, opt_default, opt_help = opt.dest, opt.default, opt.help
    if not opt_help:
        sys.stderr.write('WARNING: "%s" is missing help string.\n' % opt_name)
        opt_help = ""
    opt_type = None
    try:
        opt_type = OPTION_REGEX.search(str(type(opt))).group(0)
    except (ValueError, AttributeError) as err:
        sys.stderr.write("%s\n" % str(err))
        sys.exit(1)
    opt_help = u'%s (%s)' % (opt_help,
                             OPT_TYPES[opt_type])
    print('#', "\n# ".join(textwrap.wrap(opt_help, WORDWRAP_WIDTH)))
    if opt.deprecated_opts:
        for deprecated_opt in opt.deprecated_opts:
            if deprecated_opt.name:
                deprecated_group = (deprecated_opt.group if
                                    deprecated_opt.group else "DEFAULT")
                print('# Deprecated group/name - [%s]/%s' %
                      (deprecated_group,
                       deprecated_opt.name))
    try:
        if opt_default is None:
            print('#%s=<None>' % opt_name)
        elif opt_type == STROPT:
            assert(isinstance(opt_default, six.string_types))
            print('#%s=%s' % (opt_name, _sanitize_default(opt_name,
                                                          opt_default)))
        elif opt_type == BOOLOPT:
            assert(isinstance(opt_default, bool))
            print('#%s=%s' % (opt_name, str(opt_default).lower()))
        elif opt_type == INTOPT:
            assert(isinstance(opt_default, int) and
                   not isinstance(opt_default, bool))
            print('#%s=%s' % (opt_name, opt_default))
        elif opt_type == FLOATOPT:
            assert(isinstance(opt_default, float))
            print('#%s=%s' % (opt_name, opt_default))
        elif opt_type == LISTOPT:
            assert(isinstance(opt_default, list))
            print('#%s=%s' % (opt_name, ','.join(opt_default)))
        elif opt_type == DICTOPT:
            assert(isinstance(opt_default, dict))
            opt_default_strlist = [str(key) + ':' + str(value)
                                   for (key, value) in opt_default.items()]
            print('#%s=%s' % (opt_name, ','.join(opt_default_strlist)))
        elif opt_type == MULTISTROPT:
            assert(isinstance(opt_default, list))
            if not opt_default:
                opt_default = ['']
            for default in opt_default:
                print('#%s=%s' % (opt_name, default))
        print('')
    except Exception:
        sys.stderr.write('Error in option "%s"\n' % opt_name)
        sys.exit(1)


def main():
    generate(sys.argv[1:])

if __name__ == '__main__':
    main()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneclient.v3 import client as ks_client
from oslo.config import cfg

from mistral import context

CONF = cfg.CONF


def client():
    ctx = context.ctx()
    auth_url = CONF.keystone_authtoken.auth_uri

    keystone = ks_client.Client(username=ctx.user_name,
                                token=ctx.auth_token,
                                tenant_id=ctx.project_id,
                                auth_url=auth_url)
    keystone.management_url = auth_url

    return keystone


def _admin_client(trust_id=None, project_name=None):
    auth_url = CONF.keystone_authtoken.auth_uri

    client = ks_client.Client(username=CONF.keystone_authtoken.admin_user,
                              password=CONF.keystone_authtoken.admin_password,
                              project_name=project_name,
                              auth_url=auth_url,
                              trust_id=trust_id)
    client.management_url = auth_url

    return client


def client_for_admin(project_name):
    return _admin_client(project_name=project_name)


def client_for_trusts(trust_id):
    return _admin_client(trust_id=trust_id)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo.config import cfg

from mistral.db import api as db_api
from mistral import expressions as expr
from mistral.openstack.common import log as logging
from mistral.services import trusts


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def evaluate_task_parameters(task, context):
    params = task['task_spec'].get('parameters', {})

    return expr.evaluate_recursively(params, context)


def prepare_tasks(tasks, context):
    for task in tasks:
        # TODO(rakhmerov): Inbound context should be a merge of outbound
        # contexts of task dependencies, if any.
        task['in_context'] = context
        task['parameters'] = evaluate_task_parameters(task, context)

        db_api.task_update(task['id'],
                           {'in_context': task['in_context'],
                            'parameters': task['parameters']})


def get_task_output(task, result):
    publish_transformer = task['task_spec'].get('publish')

    output = expr.evaluate_recursively(publish_transformer, result) or {}

    if result:
        output['task'] = {task['name']: result}

    return output


def _merge_dicts(target, src):
    for key in src:
        # TODO(nmakhotkin) Take care of the same key in both dicts
        to_merge = (key in target
                    and isinstance(target[key], dict)
                    and isinstance(src[key], dict))
        if to_merge:
            _merge_dicts(target[key], src[key])
        else:
            target[key] = src[key]
    return target


def get_outbound_context(task, output=None):
    in_context = task.get('in_context')

    out_context = in_context.copy() if in_context else {}

    if not output:
        output = task.get('output')

    if output:
        out_context = _merge_dicts(out_context, output)

    return out_context


def add_openstack_data_to_context(context, db_workbook):
    if context is None:
        context = {}

    if CONF.pecan.auth_enable:
        workbook_ctx = trusts.create_context(db_workbook)
        if workbook_ctx:
            context.update({'openstack': workbook_ctx.to_dict()})

    return context


def add_execution_to_context(context, db_execution):
    if context is None:
        context = {}

    context['__execution'] = {
        'id': db_execution['id'],
        'workbook_name': db_execution['workbook_name'],
        'task': db_execution['task']
    }

    return context



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import threading

from eventlet import corolocal

# Thread local storage.
_th_loc_storage = threading.local()


def _get_greenlet_local_storage():
    greenlet_id = corolocal.get_ident()

    greenlet_locals = getattr(_th_loc_storage, "greenlet_locals", None)

    if not greenlet_locals:
        greenlet_locals = {}
        _th_loc_storage.greenlet_locals = greenlet_locals

    if greenlet_id in greenlet_locals:
        return greenlet_locals[greenlet_id]
    else:
        return None


def has_thread_local(var_name):
    gl_storage = _get_greenlet_local_storage()
    return gl_storage and var_name in gl_storage


def get_thread_local(var_name):
    if not has_thread_local(var_name):
        return None

    return _get_greenlet_local_storage()[var_name]


def set_thread_local(var_name, val):
    if not val and has_thread_local(var_name):
        gl_storage = _get_greenlet_local_storage()

        # Delete variable from greenlet local storage.
        if gl_storage:
            del gl_storage[var_name]

        # Delete the entire greenlet local storage from thread local storage.
        if gl_storage and len(gl_storage) == 0:
            del _th_loc_storage.greenlet_locals[corolocal.get_ident()]

    if val:
        gl_storage = _get_greenlet_local_storage()
        if not gl_storage:
            gl_storage =\
                _th_loc_storage.greenlet_locals[corolocal.get_ident()] = {}

        gl_storage[var_name] = val


def log_exec(logger, level=logging.INFO):
    """Decorator for logging function execution.
        By default, target function execution is logged with INFO level.
    """

    def _decorator(func):
        def _logged(*args, **kw):
            params_repr = "[args=%s, kw=%s]" % (str(args), str(kw)) \
                if len(args) > 0 or len(kw) > 0 else ""

            func_repr = "Called method [name=%s, doc='%s', params=%s]" % \
                        (func.__name__, func.__doc__, params_repr)

            logger.log(level, func_repr)

            return func(*args, **kw)

        _logged.__doc__ = func.__doc__

        return _logged

    return _decorator



# -*- coding: utf-8 -*-
#
# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from email import parser
import mock
import testtools

from mistral.actions import std_actions as std
from mistral import exceptions as exc
from mistral.tests import base

"""
To try against a real SNMP server:

1) set LOCAL_SMPTD = True
   run debug snmpd on the local machine:
   `sudo python -m smtpd -c DebuggingServer -n localhost:25`
   Debugging server doesn't support password.

2) set REMOTE_SMPT = True
   use external SNMP (like gmail), change the configuration,
   provide actual username and password
        self.settings = {
            'host': 'smtp.gmail.com:587',
            'from': "youraccount@gmail.com",
            'password': "secret"
        }

"""

LOCAL_SMTPD = False
REMOTE_SMTP = False


class SendEmailActionTest(base.BaseTest):

    def setUp(self):
        super(SendEmailActionTest, self).setUp()
        self.params = {
            'to': ["dz@example.com, deg@example.com", "xyz@example.com"],
            'subject': "Multi word subject с русскими буквами",
            'body': "short multiline\nbody\nc русскими буквами",
        }
        self.settings = {
            'smtp_server': 'mail.example.com:25',
            'from': "bot@example.com",
        }
        self.to_addrs = ', '.join(self.params['to'])

    @testtools.skipIf(not LOCAL_SMTPD, "Setup local smtpd to run it")
    def test_send_email_real(self):
        action = std.SendEmailAction(self.params, self.settings)
        action.run()

    @testtools.skipIf(not REMOTE_SMTP, "Configure Remote SMTP to run it")
    def test_with_password_real(self):
        self.params['to'] = ["dz@stackstorm.com"]
        self.settings = {
            'smtp_server': 'smtp.gmail.com:587',
            'from': "username@gmail.com",
            'password': 'secret'
        }

        action = std.SendEmailAction(self.params, self.settings)

        action.run()

    @mock.patch('smtplib.SMTP')
    def test_send_email(self, smtp):
        action = std.SendEmailAction(self.params, self.settings)

        action.run()

        smtp.assert_called_once_with(self.settings['smtp_server'])

        sendmail = smtp.return_value.sendmail

        self.assertTrue(sendmail.called, "should call sendmail")
        self.assertEqual(
            sendmail.call_args[1]['from_addr'], self.settings['from'])
        self.assertEqual(
            sendmail.call_args[1]['to_addrs'], self.to_addrs)

        message = parser.Parser().parsestr(sendmail.call_args[1]['msg'])

        self.assertEqual(self.settings['from'], message['from'])
        self.assertEqual(self.to_addrs, message['to'])
        self.assertEqual(self.params['subject'], message['subject'])
        self.assertEqual(self.params['body'], message.get_payload())

    @mock.patch('smtplib.SMTP')
    def test_with_password(self, smtp):
        self.settings['password'] = "secret"

        action = std.SendEmailAction(self.params, self.settings)

        action.run()

        smtpmock = smtp.return_value
        calls = [mock.call.ehlo(), mock.call.starttls(), mock.call.ehlo(),
                 mock.call.login(self.settings['from'],
                                 self.settings['password'])]

        smtpmock.assert_has_calls(calls)
        self.assertTrue(smtpmock.sendmail.called, "should call sendmail")

    @mock.patch('mistral.actions.std_actions.LOG')
    def test_exception(self, log):
        self.params['smtp_server'] = "wrong host"

        action = std.SendEmailAction(self.params, self.settings)

        try:
            action.run()
        except exc.ActionException:
            pass
        else:
            self.assertFalse("Must throw exception.")



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo.config import cfg

from mistral.actions import action_factory as a_f
from mistral.db import api as db_api
from mistral.engine import executor
from mistral.engine import states
from mistral import exceptions as exc
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)
WORKFLOW_TRACE = logging.getLogger(cfg.CONF.workflow_trace_log_name)


class DefaultExecutor(executor.Executor):

    def _do_task_action(self, task):
        """Executes the action defined by the task and return result.

        :param task: a task definition
        :type task: dict
        """
        LOG.info("Starting task action [task_id=%s, "
                 "action='%s', action_spec='%s'" %
                 (task['id'], task['task_spec']['action'],
                  task['action_spec']))

        action = a_f.create_action(task)

        if action.is_sync():
            try:
                state, result = states.SUCCESS, action.run()
            except exc.ActionException:
                state, result = states.ERROR, None

            self.engine.convey_task_result(task['id'],
                                           state, result)
        else:
            try:
                action.run()
            except exc.ActionException:
                self.engine.convey_task_result(task['id'],
                                               states.ERROR, None)

    def _handle_task_error(self, task, exception):
        """Handle exception from the task execution.

        :param task: the task corresponding to the exception
        :type task: dict
        :param exception: an exception thrown during the execution of the task
        :type exception: Exception
        """
        try:
            db_api.start_tx()
            try:
                db_api.execution_update(task['execution_id'],
                                        {'state': states.ERROR})
                db_api.task_update(task['id'],
                                   {'state': states.ERROR})
                db_api.commit_tx()
            finally:
                db_api.end_tx()
        except Exception as e:
            LOG.exception(e)

    def handle_task(self, cntx, **kwargs):
        """Handle the execution of the workbook task.

        :param cntx: a request context dict
        :type cntx: MistralContext
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        """
        try:
            task = kwargs.get('task', None)
            if not task:
                raise Exception('No task is provided to the executor.')

            LOG.info("Received a task: %s" % task)

            db_task = db_api.task_get(task['id'])
            db_exec = db_api.execution_get(task['execution_id'])

            if not db_exec or not db_task:
                return

            if db_exec['state'] != states.RUNNING or \
                    db_task['state'] != states.IDLE:
                return

            # Update the state to running before performing action. The
            # do_task_action assigns state to the task which is the appropriate
            # value to preserve.

            WORKFLOW_TRACE.info("Task '%s' [%s -> %s]" % (db_task['name'],
                                                          db_task['state'],
                                                          states.RUNNING))

            db_api.task_update(task['id'],
                               {'state': states.RUNNING})

            self._do_task_action(db_task)
        except Exception as e:
            LOG.exception(e)
            self._handle_task_error(task, e)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import eventlet
import mock
from oslo.config import cfg

eventlet.monkey_patch()

from mistral.actions import std_actions
from mistral.db import api as db_api
from mistral import engine
from mistral.engine import executor
from mistral.engine import states
from mistral.openstack.common import importutils
from mistral.openstack.common import log as logging
from mistral.tests import base


# We need to make sure that all configuration properties are registered.
importutils.import_module("mistral.config")
LOG = logging.getLogger(__name__)

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WORKBOOK_NAME = 'my_workbook'
TASK_NAME = 'create-vms'

SAMPLE_WORKBOOK = {
    'id': str(uuid.uuid4()),
    'name': WORKBOOK_NAME,
    'description': 'my description',
    'definition': base.get_resource("test_rest.yaml"),
    'tags': [],
    'scope': 'public',
    'updated_at': None,
    'project_id': '123',
    'trust_id': '1234'
}

SAMPLE_EXECUTION = {
    'id': str(uuid.uuid4()),
    'workbook_name': WORKBOOK_NAME,
    'task': TASK_NAME,
    'state': states.RUNNING,
    'updated_at': None,
    'context': None
}

SAMPLE_TASK = {
    'name': TASK_NAME,
    'workbook_name': WORKBOOK_NAME,
    'action_spec': {
        'name': 'my-action',
        'class': 'std.http',
        'base-parameters': {
            'url': 'http://localhost:8989/v1/workbooks',
            'method': 'GET'
        },
        'namespace': 'MyRest'
    },
    'in_context': {},
    'parameters': {},
    'task_spec': {
        'action': 'MyRest.my-action',
        'name': TASK_NAME},
    'requires': [],
    'state': states.IDLE}

SAMPLE_CONTEXT = {
    'user': 'admin',
    'tenant': 'mistral'
}


class TestExecutor(base.DbTestCase):
    def __init__(self, *args, **kwargs):
        super(TestExecutor, self).__init__(*args, **kwargs)
        self.transport = base.get_fake_transport()

    @mock.patch.object(
        executor.ExecutorClient, 'handle_task',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_handle_task))
    @mock.patch.object(
        std_actions.HTTPAction, 'run', mock.MagicMock(return_value={}))
    @mock.patch.object(
        engine.EngineClient, 'convey_task_result',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_task_result))
    def test_handle_task(self):
        # Create a new workbook.
        workbook = db_api.workbook_create(SAMPLE_WORKBOOK)
        self.assertIsInstance(workbook, dict)

        # Create a new execution.
        execution = db_api.execution_create(SAMPLE_EXECUTION['workbook_name'],
                                            SAMPLE_EXECUTION)
        self.assertIsInstance(execution, dict)

        # Create a new task.
        SAMPLE_TASK['execution_id'] = execution['id']
        task = db_api.task_create(SAMPLE_TASK['execution_id'],
                                  SAMPLE_TASK)
        self.assertIsInstance(task, dict)
        self.assertIn('id', task)

        # Send the task request to the Executor.
        ex_client = executor.ExecutorClient(self.transport)
        ex_client.handle_task(SAMPLE_CONTEXT, task=task)

        # Check task execution state.
        db_task = db_api.task_get(task['id'])
        self.assertEqual(db_task['state'], states.SUCCESS)



# -*- coding: utf-8 -*-
#
# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import json

from mistral.actions import action_factory as a_f
from mistral.actions import std_actions as std
from mistral.engine import data_flow
from mistral import exceptions
from mistral.openstack.common import log as logging
from mistral.tests import base


LOG = logging.getLogger(__name__)

DB_TASK = {
    'task_spec': {
        'name': 'my_task',
        'action': 'std.http'
    },
    'name': 'my_task',
    'workbook_name': 'my_workbook',
    'execution_id': '123',
    'id': '123',
    'in_context': {},
    'tags': ['deployment', 'test'],
    'parameters': {
        'url': 'http://some.url',
        'params': {
            'param1': 'val1',
            'param2': 'val2'
        },
        'method': 'POST',
        'headers': {
            'content-type': 'text/json'
        },
        'body': {
            'my_object': {
                'prop1': 'val1',
                'prop2': 'val2'
            }
        }
    }
}

DB_TASK_ADHOC = {
    'task_spec': {
        'name': 'my_task',
        'action': 'my_namespace.my_action'
    },
    'action_spec': {
        'name': 'my_action',
        'namespace': 'my_namespace',
        'class': 'std.echo',
        'base-parameters': {
            'output': '{$.first} and {$.second}'
        },
        'parameters': ['first', 'second'],
        'output': {
            'res': '{$}'
        }
    },
    'name': 'my_task',
    'workbook_name': 'my_workbook',
    'execution_id': '123',
    'id': '123',
    'in_context': {},
    'tags': ['deployment', 'test'],
    'parameters': {
        'first': 'Tango',
        'second': 'Cash'
    }
}


class ActionFactoryTest(base.BaseTest):
    def test_register_standard_actions(self):
        namespaces = a_f.get_registered_namespaces()

        self.assertEqual(1, len(namespaces))
        self.assertIn("std", namespaces)

        std_ns = namespaces["std"]

        self.assertEqual(5, len(std_ns))

        self.assertTrue(std_ns.contains_action_name("echo"))
        self.assertTrue(std_ns.contains_action_name("http"))
        self.assertTrue(std_ns.contains_action_name("mistral_http"))
        self.assertTrue(std_ns.contains_action_name("email"))

        self.assertEqual(std.EchoAction, std_ns.get_action_class("echo"))
        self.assertEqual(std.HTTPAction, std_ns.get_action_class("http"))
        self.assertEqual(std.MistralHTTPAction,
                         std_ns.get_action_class("mistral_http"))
        self.assertEqual(std.SendEmailAction,
                         std_ns.get_action_class("email"))

    def test_get_action_class(self):
        self.assertEqual(std.EchoAction, a_f.get_action_class("std.echo"))
        self.assertEqual(std.HTTPAction, a_f.get_action_class("std.http"))
        self.assertEqual(std.SendEmailAction,
                         a_f.get_action_class("std.email"))

    def test_get_action_class_failure(self):
        exc = self.assertRaises(exceptions.ActionException,
                                a_f.get_action_class, 'echo')
        self.assertIn('Invalid action name', exc.message)

    def test_create_http_action(self):
        action = a_f.create_action(DB_TASK)

        self.assertIsNotNone(action)

        task_params = DB_TASK['parameters']

        self.assertEqual(task_params['url'], action.url)
        self.assertDictEqual(task_params['params'], action.params)
        self.assertEqual(task_params['method'], action.method)
        self.assertEqual(task_params['headers'], action.headers)
        self.assertDictEqual(task_params['body'], json.loads(action.body))

    def test_create_mistral_http_action(self):
        db_task = copy.copy(DB_TASK)
        db_task['task_spec']['action'] = 'std.mistral_http'

        action = a_f.create_action(db_task)

        self.assertIsNotNone(action)

        task_params = db_task['parameters']

        self.assertEqual(task_params['url'], action.url)
        self.assertDictEqual(task_params['params'], action.params)
        self.assertEqual(task_params['method'], action.method)
        self.assertEqual(task_params['headers'], action.headers)
        self.assertDictEqual(task_params['body'], json.loads(action.body))

        headers = task_params['headers']

        self.assertEqual(db_task['workbook_name'],
                         headers['Mistral-Workbook-Name'])
        self.assertEqual(db_task['execution_id'],
                         headers['Mistral-Execution-Id'])
        self.assertEqual(db_task['id'],
                         headers['Mistral-Task-Id'])

    def test_create_adhoc_action_with_openstack_context(self):
        db_task = copy.copy(DB_TASK_ADHOC)
        db_task['action_spec']['output'] = {'res': '{$}'}
        db_task['in_context'].update({
            'openstack': {
                'auth_token': '123',
                'project_id': '321'
            }
        })
        base_parameters = db_task['action_spec']['base-parameters']
        base_parameters['output'] = ("{$.openstack.auth_token}"
                                     "{$.openstack.project_id}")

        action = a_f.create_action(db_task)

        self.assertEqual({'res': "123321"}, action.run())

    def test_create_adhoc_action_no_openstack_context(self):
        db_task = copy.copy(DB_TASK_ADHOC)
        db_task['action_spec']['output'] = {'res': '{$}'}
        db_task['in_context'].update({
            'openstack': None
        })
        base_parameters = db_task['action_spec']['base-parameters']
        base_parameters['output'] = "$.openstack.auth_token"

        action = a_f.create_action(db_task)

        self.assertEqual({'res': "$.openstack.auth_token"}, action.run())

    def test_create_no_adhoc_action_with_openstack_context(self):
        db_task = copy.copy(DB_TASK)
        db_task['task_spec']['action'] = 'std.http'
        db_task['in_context'].update({
            'openstack': {
                'auth_token': '123',
                'project_id': '321'
            }
        })
        ## In case of no-adhoc action we should evaluate task parameters
        ## to see what we need.
        task_spec = db_task['task_spec']
        task_spec['parameters'] = {
            'url': "http://some/{$.openstack.project_id}/servers",
        }

        db_task['parameters'] = data_flow.evaluate_task_parameters(
            db_task, db_task['in_context'])

        action = a_f.create_action(db_task)

        self.assertEqual("http://some/321/servers", action.url)

    def test_get_ssh_action(self):
        db_task = copy.copy(DB_TASK)
        db_task['task_spec'] = {
            'name': 'my_task',
            'action': 'std.ssh',
            'parameters': {
                'host': '10.0.0.1',
                'cmd': 'ls -l',
                'username': '$.ssh_username',
                'password': '$.ssh_password'
            }
        }
        db_task['parameters'] = {'host': '10.0.0.1',
                                 'cmd': 'ls -l',
                                 'username': 'ubuntu',
                                 'password': 'ubuntu_password'}

        action = a_f.create_action(db_task)

        self.assertEqual("ubuntu", action.username)
        self.assertEqual("ubuntu_password", action.password)
        self.assertEqual("ls -l", action.cmd)
        self.assertEqual("10.0.0.1", action.host)

    def test_adhoc_echo_action(self):
        db_task = copy.copy(DB_TASK_ADHOC)
        action_spec = db_task['action_spec']

        # With dic-like output formatter.
        action = a_f.create_action(db_task)

        self.assertDictEqual({'res': 'Tango and Cash'}, action.run())

        # With list-like output formatter.
        action_spec['output'] = ['$', '$']

        action = a_f.create_action(db_task)

        self.assertListEqual(['Tango and Cash', 'Tango and Cash'],
                             action.run())

        # With single-object output formatter.
        action_spec['output'] = "'{$}' is a cool movie!"

        action = a_f.create_action(db_task)

        self.assertEqual("'Tango and Cash' is a cool movie!", action.run())



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Access Control API server."""

from keystoneclient.middleware import auth_token
from oslo.config import cfg


_ENFORCER = None


def setup(app):
    if cfg.CONF.pecan.auth_enable:
        return auth_token.AuthProtocol(app, dict(cfg.CONF.keystone_authtoken))
    else:
        return app


def get_limited_to(headers):
    """Return the user and project the request should be limited to.

    :param headers: HTTP headers dictionary
    :return: A tuple of (user, project), set to None if there's no limit on
    one of these.

    """
    return headers.get('X-User-Id'), headers.get('X-Project-Id')


def get_limited_to_project(headers):
    """Return the project the request should be limited to.

    :param headers: HTTP headers dictionary
    :return: A project, or None if there's no limit on it.

    """
    return get_limited_to(headers)[1]



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pecan

from mistral.db import api as db_api
from mistral.openstack.common import log as logging
from mistral.services import scheduler
from mistral.utils import rest_utils


LOG = logging.getLogger(__name__)


class WorkbookDefinitionController(pecan.rest.RestController):
    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose()
    def get(self, workbook_name):
        """Return the workbook definition."""
        LOG.debug("Fetch workbook definition [workbook_name=%s]" %
                  workbook_name)
        return db_api.workbook_definition_get(workbook_name)

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose(content_type="text/plain")
    def put(self, workbook_name):
        """Update the workbook's definition."""
        text = pecan.request.text

        LOG.debug("Update workbook definition [workbook_name=%s, text=%s]" %
                  (workbook_name, text))
        wb = db_api.workbook_definition_put(workbook_name, text)
        scheduler.create_associated_triggers(wb)
        return wb['definition']



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from oslo.config import cfg
import pecan
import pecan.testing
from webtest import app as webtest_app

from mistral.tests import base


# Disable authentication for functional tests.
cfg.CONF.set_default('auth_enable', False, group='pecan')


__all__ = ['FunctionalTest']


class FunctionalTest(base.DbTestCase):
    """Used for functional tests where you need to test your
    literal application and its integration with the framework.
    """

    def setUp(self):
        super(FunctionalTest, self).setUp()

        pecan_opts = cfg.CONF.pecan

        self.app = pecan.testing.load_test_app({
            'app': {
                'root': pecan_opts.root,
                'modules': pecan_opts.modules,
                'debug': pecan_opts.debug,
                'auth_enable': False
            }
        })
        self.addCleanup(pecan.set_config, {}, overwrite=True)
        self.addCleanup(cfg.CONF.set_default, 'auth_enable', False,
                        group='pecan')

        # make sure the api get the correct context.
        self.patch_ctx = mock.patch('mistral.context.context_from_headers')
        self.mock_ctx = self.patch_ctx.start()
        self.mock_ctx.return_value = self.ctx
        self.addCleanup(self.patch_ctx.stop)

    def assertNotFound(self, url):
        try:
            self.app.get(url, headers={'Accept': 'application/json'})
        except webtest_app.AppError as error:
            self.assertIn('Bad response: 404 Not Found', str(error))
            return
        self.fail('Expected 404 Not found but got OK')

    def assertUnauthorized(self, url):
        try:
            self.app.get(url, headers={'Accept': 'application/json'})
        except webtest_app.AppError as error:
            self.assertIn('Bad response: 401 Unauthorized', str(error))
            return
        self.fail('Expected 401 Unauthorized but got OK')



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from croniter import croniter
import datetime
from mistral.db import api as db_api
from mistral import dsl_parser as parser


def get_next_triggers():
    return db_api.get_next_triggers(datetime.datetime.now() +
                                    datetime.timedelta(0, 2))


def set_next_execution_time(trigger):
    base = trigger['next_execution_time']
    cron = croniter(trigger['pattern'], base)

    return db_api.trigger_update(trigger['id'], {
        'next_execution_time': cron.get_next(datetime.datetime)
    })


def _get_next_execution_time(pattern, start_time):
    return croniter(pattern, start_time).get_next(datetime.datetime)


def create_trigger(name, pattern, workbook_name, start_time=None):
    if not start_time:
        start_time = datetime.datetime.now()

    return db_api.trigger_create({
        "name": name,
        "pattern": pattern,
        "next_execution_time": _get_next_execution_time(pattern, start_time),
        "workbook_name": workbook_name
    })


def create_associated_triggers(db_workbook):
    if not db_workbook['definition']:
        return

    workbook = parser.get_workbook(db_workbook['definition'])
    triggers = workbook.get_triggers()

    # Prepare all triggers data in advance to make db transaction shorter.
    db_triggers = []

    for e in triggers:
        pattern = e['parameters']['cron-pattern']
        next_time = _get_next_execution_time(pattern, datetime.datetime.now())
        db_triggers.append({
            "name": e['name'],
            "pattern": pattern,
            "next_execution_time": next_time,
            "workbook_name": db_workbook['name']
        })

    db_api.start_tx()

    try:
        for e in db_triggers:
            db_api.trigger_create(e)

        db_api.commit_tx()
    finally:
        db_api.end_tx()



# -*- coding: utf-8 -*-
#
# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from email.mime import text
import json
import requests
import smtplib

from mistral.actions import base
from mistral import exceptions as exc
from mistral import expressions as expr
from mistral.openstack.common import log as logging
from mistral.utils import ssh_utils


LOG = logging.getLogger(__name__)


class EchoAction(base.Action):
    """Echo action.

    This action just returns a configured value as a result without doing
    anything else. The value of such action implementation is that it
    can be used in development (for testing), demonstration and designing
    of workflows themselves where echo action can play the role of temporary
    stub.
    """

    def __init__(self, output):
        self.output = output

    def run(self):
        LOG.info('Running echo action [output=%s]' % self.output)

        return self.output

    def test(self):
        return 'Echo'


class HTTPAction(base.Action):
    """Constructs an HTTP action.

    :param url: URL for the new HTTP request.
    :param method: (optional, 'GET' by default) method for the new HTTP
        request.
    :param params: (optional) Dictionary or bytes to be sent in the
        query string for the HTTP request.
    :param body: (optional) Dictionary, bytes, or file-like object to send
        in the body of the HTTP request.
    :param headers: (optional) Dictionary of HTTP Headers to send with
        the HTTP request.
    :param cookies: (optional) Dict or CookieJar object to send with
        the HTTP request.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom
        HTTP Auth.
    :param timeout: (optional) Float describing the timeout of the request
        in seconds.
    :param allow_redirects: (optional) Boolean. Set to True if POST/PUT/DELETE
        redirect following is allowed.
    :param proxies: (optional) Dictionary mapping protocol to the URL of
        the proxy.
    """

    def __init__(self,
                 url,
                 method="GET",
                 params=None,
                 body=None,
                 headers=None,
                 cookies=None,
                 auth=None,
                 timeout=None,
                 allow_redirects=None,
                 proxies=None):
        self.url = url
        self.method = method
        self.params = params
        self.body = json.dumps(body) if isinstance(body, dict) else body
        self.headers = headers
        self.cookies = cookies
        self.auth = auth
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.proxies = proxies

    def run(self):
        LOG.info("Running HTTP action "
                 "[url=%s, method=%s, params=%s, body=%s, headers=%s,"
                 " cookies=%s, auth=%s, timeout=%s, allow_redirects=%s,"
                 " proxies=%s]" %
                 (self.url,
                  self.method,
                  self.params,
                  self.body,
                  self.headers,
                  self.cookies,
                  self.auth,
                  self.timeout,
                  self.allow_redirects,
                  self.proxies))

        try:
            resp = requests.request(
                self.method,
                self.url,
                params=self.params,
                data=self.body,
                headers=self.headers,
                cookies=self.cookies,
                auth=self.auth,
                timeout=self.timeout,
                allow_redirects=self.allow_redirects,
                proxies=self.proxies
            )
        except Exception as e:
            raise exc.ActionException("Failed to send HTTP request: %s" % e)

        LOG.info("HTTP action response:\n%s\n%s" %
                 (resp.status_code, resp.content))

        # TODO(everyone): Not sure we need to have this check here in base HTTP
        #                 action.
        if resp.status_code not in range(200, 307):
            raise exc.ActionException("Received error HTTP code: %s" %
                                      resp.status_code)

        # Construct all important resp data in readable structure.
        headers = dict(resp.headers.items())
        status = resp.status_code
        try:
            content = resp.json()
        except Exception as e:
            LOG.debug("HTTP action response is not json.")
            content = resp.content

        return {'content': content, 'headers': headers, 'status': status}


class MistralHTTPAction(HTTPAction):
    def __init__(self,
                 action_context,
                 url,
                 method="GET",
                 params=None,
                 body=None,
                 headers=None,
                 cookies=None,
                 auth=None,
                 timeout=None,
                 allow_redirects=None,
                 proxies=None):
        headers.update({
            'Mistral-Workbook-Name': action_context['workbook_name'],
            'Mistral-Execution-Id': action_context['execution_id'],
            'Mistral-Task-Id': action_context['task_id'],
        })

        super(MistralHTTPAction, self).__init__(
            url,
            method,
            params,
            body,
            headers,
            cookies,
            auth,
            timeout,
            allow_redirects,
            proxies,
        )

    def is_sync(self):
        return False


class SendEmailAction(base.Action):
    def __init__(self, params, settings):
        # TODO(dzimine): validate parameters

        # Task invocation parameters.
        self.to = ', '.join(params['to'])
        self.subject = params['subject']
        self.body = params['body']

        # Action provider settings.
        self.smtp_server = settings['smtp_server']
        self.sender = settings['from']
        self.password = settings['password'] \
            if 'password' in settings else None

    def run(self):
        LOG.info("Sending email message "
                 "[from=%s, to=%s, subject=%s, using smtp=%s, body=%s...]" %
                 (self.sender, self.to, self.subject,
                  self.smtp_server, self.body[:128]))

        # TODO(dzimine): handle utf-8, http://stackoverflow.com/a/14506784
        message = text.MIMEText(self.body)
        message['Subject'] = self.subject
        message['From'] = self.sender
        message['To'] = self.to

        try:
            s = smtplib.SMTP(self.smtp_server)

            if self.password is not None:
                # Sequence to request TLS connection and log in (RFC-2487).
                s.ehlo()
                s.starttls()
                s.ehlo()
                s.login(self.sender, self.password)

            s.sendmail(from_addr=self.sender,
                       to_addrs=self.to,
                       msg=message.as_string())
        except (smtplib.SMTPException, IOError) as e:
            raise exc.ActionException("Failed to send an email message: %s"
                                      % e)

    def test(self):
        # Just logging the operation since this action is not supposed
        # to return a result.
        LOG.info("Sending email message "
                 "[from=%s, to=%s, subject=%s, using smtp=%s, body=%s...]" %
                 (self.sender, self.to, self.subject,
                  self.smtp_server, self.body[:128]))


class SSHAction(base.Action):
    def __init__(self, cmd, host, username, password):
        self.cmd = cmd
        self.host = host
        self.username = username
        self.password = password

    def run(self):
        def raise_exc(parent_exc=None):
            message = ("Failed to execute ssh cmd "
                       "'%s' on %s" % (self.cmd, self.host))
            if parent_exc:
                message += "\nException: %s" % str(parent_exc)
            raise exc.ActionException(message)

        try:
            status_code, result = ssh_utils.execute_command(self.cmd,
                                                            self.host,
                                                            self.username,
                                                            self.password)
            if status_code > 0:
                return raise_exc()

            return result
        except Exception as e:
            return raise_exc(parent_exc=e)


class AdHocAction(base.Action):
    def __init__(self, action_context,
                 base_action_cls, action_spec, **params):
        self.base_action_cls = base_action_cls
        self.action_spec = action_spec

        base_params = self._convert_params(params)

        if action_context:
            base_params['action_context'] = action_context

        self.base_action = base_action_cls(**base_params)

    def _convert_params(self, params):
        base_params_spec = self.action_spec.base_parameters

        if not base_params_spec:
            return {}

        return expr.evaluate_recursively(base_params_spec, params)

    def _convert_result(self, result):
        transformer = self.action_spec.output

        if not transformer:
            return result

        # Use base action result as a context for evaluating expressions.
        return expr.evaluate_recursively(transformer, result)

    def is_sync(self):
        return self.base_action.is_sync()

    def run(self):
        try:
            return self._convert_result(self.base_action.run())
        except Exception as e:
            raise exc.ActionException("Failed to convert result in action %s,"
                                      "[Exception = %s"
                                      % (self.action_spec.name, e))

    def test(self):
        return self._convert_result(self.base_action.test())




# Copyright 2013 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import time

from tempest import clients
from tempest.common import rest_client
from tempest import config
from tempest import exceptions
import tempest.test

CONF = config.CONF


class MistralClient(rest_client.RestClient):

    def __init__(self, auth_provider):
        super(MistralClient, self).__init__(auth_provider)

        self.service = 'workflow_service'
        self.endpoint_url = 'publicURL'

    def get_list_obj(self, name):
        resp, body = self.get(name)
        return resp, json.loads(body)

    def create_obj(self, path, name):
        post_body = '{"name": "%s"}' % name
        resp, body = self.post(path, post_body)
        return resp, json.loads(body)

    def delete_obj(self, path, name):
        return self.delete('{path}/{name}'.format(path=path, name=name))

    def update_obj(self, path, name):
        post_body = '{"name": "%s"}' % (name + 'updated')
        resp, body = self.put('{path}/{name}'.format(path=path, name=name),
                              post_body)
        return resp, json.loads(body)

    def get_workbook_definition(self, name):
        headers = {'X-Auth-Token': self.auth_provider.get_token()}
        return self.get('workbooks/{name}/definition'.format(name=name),
                        headers)

    def upload_workbook_definition(self, name):
        headers = {'Content-Type': 'text/plain',
                   'X-Auth-Token': self.auth_provider.get_token()}

        __location = os.path.realpath(os.path.join(os.getcwd(),
                                                   os.path.dirname(__file__)))

        file = open(os.path.join(__location, 'v1/demo.yaml'), 'rb').read()
        return self.put('workbooks/{name}/definition'.format(name=name),
                        file, headers)

    def create_execution(self, workbook_name, body):
        return self.post('workbooks/{name}/executions'.format(
            name=workbook_name), json.dumps(body))

    def get_execution(self, workbook_name, execution_id):
        return self.get('/workbooks/{name}/executions/{execution}'.format(
            name=workbook_name, execution=execution_id))

    def update_execution(self, workbook_name, execution_id, put_body):
        return self.put('/workbooks/{name}/executions/{execution}'.format(
            name=workbook_name, execution=execution_id), json.dumps(put_body))

    def get_tasks_list(self, workbook_name, execution_id):
        resp, body = self.get(
            '/workbooks/{name}/executions/{execution}/tasks'.format(
                name=workbook_name,
                execution=execution_id))

        return resp, json.loads(body)['tasks']

    def get_task(self, workbook_name, execution_id, task_id):
        resp, body = self.get(
            '/workbooks/{name}/executions/{execution}/tasks/{task}'.format(
                name=workbook_name,
                execution=execution_id,
                task=task_id))

        return resp, json.loads(body)

    def update_task(self, workbook_name, execution_id, task_id, put_body):
        resp, body = self.put(
            '/workbooks/{name}/executions/{execution}/tasks/{task}'.format(
                name=workbook_name,
                execution=execution_id,
                task=task_id), json.dumps(put_body))

        return resp, json.loads(body)


class TestCase(tempest.test.BaseTestCase):

    @classmethod
    def setUpClass(cls):
        """This method allows to initialize authentication before
        each test case and define parameters of Mistral API Service.
        """
        super(TestCase, cls).setUpClass()

        cls.mgr = clients.Manager()
        cls.client = MistralClient(cls.mgr.auth_provider)

    def setUp(self):
        super(TestCase, self).setUp()
        self.obj = []

    def tearDown(self):
        super(TestCase, self).tearDown()

        for i in self.obj:
            try:
                self.client.delete_obj(i[0], i[1])
            except exceptions.NotFound:
                pass


class TestCaseAdvanced(TestCase):

    @classmethod
    def setUpClass(cls):

        super(TestCaseAdvanced, cls).setUpClass()

        cls.server_client = cls.mgr.servers_client

        cls.image_ref = CONF.compute.image_ref
        cls.flavor_ref = CONF.compute.flavor_ref

    def setUp(self):
        super(TestCaseAdvanced, self).setUp()

        self.server_ids = []

        self.client.create_obj('workbooks', 'test123')
        self.obj.append(['workbooks', 'test123'])
        self.client.upload_workbook_definition('test123')

    def tearDown(self):
        super(TestCaseAdvanced, self).tearDown()

        for server_id in self.server_ids:
            try:
                self.server_client.delete_server(server_id)
                self.server_client.wait_for_server_termination(server_id)
            except exceptions.NotFound:
                pass

    def _create_execution(self, workbook_name, server_name):

        nova_url = "/".join(self.server_client.base_url.split('/')[:-1])

        context = {
            "server_name": server_name,
            "nova_url": nova_url,
            "image_id": self.image_ref,
            "flavor_id": self.flavor_ref
        }

        post_body = {
            "workbook_name": workbook_name,
            "task": "create-vm",
            "context": json.dumps(context)
        }

        resp, body = self.client.create_execution(workbook_name, post_body)

        while not self.server_client.list_servers()[1]['servers']:
            time.sleep(2)
        for server in self.server_client.list_servers()[1]['servers']:
            if server['name'] == server_name:
                self.server_ids.append(server['id'])

        return resp, json.loads(body)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.openstack.common import jsonutils
from mistral.tests.api import base


class TestRootController(base.FunctionalTest):

    def test_index(self):
        resp = self.app.get('/', headers={'Accept': 'application/json'})

        self.assertEqual(resp.status_int, 200)

        data = jsonutils.loads(resp.body.decode())

        self.assertEqual(data[0]['id'], 'v1.0')
        self.assertEqual(data[0]['status'], 'CURRENT')
        self.assertEqual(data[0]['link'], {'href': 'http://localhost/v1',
                                           'target': 'v1'})



# -*- coding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo.config import cfg
from oslo import messaging

from mistral import context as auth_context
from mistral import engine
from mistral.engine import executor
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)


class DefaultEngine(engine.Engine):
    def _notify_task_executors(self, tasks):
        # TODO(m4dcoder): Use a pool for transport and client
        if not self.transport:
            self.transport = messaging.get_transport(cfg.CONF)
        exctr = executor.ExecutorClient(self.transport)
        for task in tasks:
            LOG.info("Submitted task for execution: '%s'" % task)
            exctr.handle_task(auth_context.ctx(), task=task)

    def _run_tasks(self, tasks):
        # TODO(rakhmerov):
        # This call outside of DB transaction creates a window
        # when the engine may crash and DB will not be consistent with
        # the task message queue state. Need to figure out the best
        # solution to recover from this situation.
        # However, making this call in DB transaction is really bad
        # since it makes transaction much longer in time and under load
        # may overload DB with open transactions.
        self._notify_task_executors(tasks)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers import resource
from mistral.db import api as db_api
from mistral.openstack.common import log as logging
from mistral.utils import rest_utils

LOG = logging.getLogger(__name__)


class Event(resource.Resource):
    """Event descriptor."""

    @classmethod
    def sample(cls):
        return cls()


class TaskEvent(Event):
    type = "TASK_STATE"
    task = wtypes.text


class ExecutionEvent(Event):
    type = "EXECUTION_STATE"
    workbook_name = wtypes.text


class Listener(resource.Resource):
    """Listener resource."""

    id = wtypes.text
    description = wtypes.text
    workbook_name = wtypes.text
    webhook = wtypes.text
    events = [Event]

    @classmethod
    def sample(cls):
        return cls(id='1234',
                   workbook_name='flow',
                   description='listener for my flow',
                   webhook='http://example.com/here',
                   events=[Event.sample()])


class Listeners(resource.Resource):
    """A collection of Listener resources."""

    listeners = [Listener]

    @classmethod
    def sample(cls):
        return cls(listeners=[Listener.sample()])


class ListenersController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Listener, wtypes.text, wtypes.text)
    def get(self, workbook_name, id):
        """Return the specified listener."""
        LOG.debug("Fetch listener [workbook_name=%s, id=%s]" %
                  (workbook_name, id))

        values = db_api.listener_get(workbook_name, id)
        return Listener.from_dict(values)

    @wsme_pecan.wsexpose(Listener, wtypes.text, wtypes.text, body=Listener)
    def put(self, workbook_name, id, listener):
        """Update the specified listener."""
        LOG.debug("Update listener [workbook_name=%s, id=%s, listener=%s]" %
                  (workbook_name, id, listener))

        values = db_api.listener_update(workbook_name, id, listener.to_dict())

        return Listener.from_dict(values)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Listener, wtypes.text, body=Listener, status_code=201)
    def post(self, workbook_name, listener):
        """Create a new listener."""
        LOG.debug("Create listener [workbook_name=%s, listener=%s]" %
                  (workbook_name, listener))

        values = db_api.listener_create(workbook_name, listener.to_dict())

        return Listener.from_dict(values)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, wtypes.text, status_code=204)
    def delete(self, workbook_name, id):
        """Delete the specified listener."""
        LOG.debug("Delete listener [workbook_name=%s, id=%s]" %
                  (workbook_name, id))

        db_api.listener_delete(workbook_name, id)

    @wsme_pecan.wsexpose(Listeners, wtypes.text)
    def get_all(self, workbook_name):
        """Return all listeners."""
        LOG.debug("Fetch listeners [workbook_name=%s]" % workbook_name)

        listeners = [Listener.from_dict(values)
                     for values in db_api.listeners_get(workbook_name)]

        return Listeners(listeners=listeners)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import copy

from oslo.config import cfg


database_opts = [
    cfg.StrOpt('sqlite_db',
               deprecated_group='DEFAULT',
               default='mistral.sqlite',
               help='The file name to use with SQLite'),
    cfg.BoolOpt('sqlite_synchronous',
                deprecated_group='DEFAULT',
                default=True,
                help='If True, SQLite uses synchronous mode'),
    cfg.StrOpt('backend',
               default='sqlalchemy',
               deprecated_name='db_backend',
               deprecated_group='DEFAULT',
               help='The backend to use for db'),
    cfg.StrOpt('connection',
               help='The SQLAlchemy connection string used to connect to the '
                    'database',
               secret=True,
               deprecated_opts=[cfg.DeprecatedOpt('sql_connection',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('sql_connection',
                                                  group='DATABASE'),
                                cfg.DeprecatedOpt('connection',
                                                  group='sql'), ]),
    cfg.StrOpt('mysql_sql_mode',
               default='TRADITIONAL',
               help='The SQL mode to be used for MySQL sessions. '
                    'This option, including the default, overrides any '
                    'server-set SQL mode. To use whatever SQL mode '
                    'is set by the server configuration, '
                    'set this to no value. Example: mysql_sql_mode='),
    cfg.IntOpt('idle_timeout',
               default=3600,
               deprecated_opts=[cfg.DeprecatedOpt('sql_idle_timeout',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('sql_idle_timeout',
                                                  group='DATABASE'),
                                cfg.DeprecatedOpt('idle_timeout',
                                                  group='sql')],
               help='Timeout before idle sql connections are reaped'),
    cfg.IntOpt('min_pool_size',
               default=1,
               deprecated_opts=[cfg.DeprecatedOpt('sql_min_pool_size',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('sql_min_pool_size',
                                                  group='DATABASE')],
               help='Minimum number of SQL connections to keep open in a '
                    'pool'),
    cfg.IntOpt('max_pool_size',
               default=None,
               deprecated_opts=[cfg.DeprecatedOpt('sql_max_pool_size',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('sql_max_pool_size',
                                                  group='DATABASE')],
               help='Maximum number of SQL connections to keep open in a '
                    'pool'),
    cfg.IntOpt('max_retries',
               default=10,
               deprecated_opts=[cfg.DeprecatedOpt('sql_max_retries',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('sql_max_retries',
                                                  group='DATABASE')],
               help='Maximum db connection retries during startup. '
                    '(setting -1 implies an infinite retry count)'),
    cfg.IntOpt('retry_interval',
               default=10,
               deprecated_opts=[cfg.DeprecatedOpt('sql_retry_interval',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('reconnect_interval',
                                                  group='DATABASE')],
               help='Interval between retries of opening a sql connection'),
    cfg.IntOpt('max_overflow',
               default=None,
               deprecated_opts=[cfg.DeprecatedOpt('sql_max_overflow',
                                                  group='DEFAULT'),
                                cfg.DeprecatedOpt('sqlalchemy_max_overflow',
                                                  group='DATABASE')],
               help='If set, use this value for max_overflow with sqlalchemy'),
    cfg.IntOpt('connection_debug',
               default=0,
               deprecated_opts=[cfg.DeprecatedOpt('sql_connection_debug',
                                                  group='DEFAULT')],
               help='Verbosity of SQL debugging information. 0=None, '
                    '100=Everything'),
    cfg.BoolOpt('connection_trace',
                default=False,
                deprecated_opts=[cfg.DeprecatedOpt('sql_connection_trace',
                                                   group='DEFAULT')],
                help='Add python stack traces to SQL as comment strings'),
    cfg.IntOpt('pool_timeout',
               default=None,
               deprecated_opts=[cfg.DeprecatedOpt('sqlalchemy_pool_timeout',
                                                  group='DATABASE')],
               help='If set, use this value for pool_timeout with sqlalchemy'),
    cfg.BoolOpt('use_db_reconnect',
                default=False,
                help='Enable the experimental use of database reconnect '
                     'on connection lost'),
    cfg.IntOpt('db_retry_interval',
               default=1,
               help='seconds between db connection retries'),
    cfg.BoolOpt('db_inc_retry_interval',
                default=True,
                help='Whether to increase interval between db connection '
                     'retries, up to db_max_retry_interval'),
    cfg.IntOpt('db_max_retry_interval',
               default=10,
               help='max seconds between db connection retries, if '
                    'db_inc_retry_interval is enabled'),
    cfg.IntOpt('db_max_retries',
               default=20,
               help='maximum db connection retries before error is raised. '
                    '(setting -1 implies an infinite retry count)'),
]

CONF = cfg.CONF
CONF.register_opts(database_opts, 'database')


def set_defaults(sql_connection, sqlite_db, max_pool_size=None,
                 max_overflow=None, pool_timeout=None):
    """Set defaults for configuration variables."""
    cfg.set_defaults(database_opts,
                     connection=sql_connection,
                     sqlite_db=sqlite_db)
    # Update the QueuePool defaults
    if max_pool_size is not None:
        cfg.set_defaults(database_opts,
                         max_pool_size=max_pool_size)
    if max_overflow is not None:
        cfg.set_defaults(database_opts,
                         max_overflow=max_overflow)
    if pool_timeout is not None:
        cfg.set_defaults(database_opts,
                         pool_timeout=pool_timeout)


def list_opts():
    """Returns a list of oslo.config options available in the library.

    The returned list includes all oslo.config options which may be registered
    at runtime by the library.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by this library.

    :returns: a list of (group_name, opts) tuples
    """
    return [('database', copy.deepcopy(database_opts))]



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


class Error(Exception):
    def __init__(self, message=None):
        super(Error, self).__init__(message)


class MistralException(Error):
    """Base Exception for the project

    To correctly use this class, inherit from it and define
    a 'message' and 'http_code' properties.
    """
    message = "An unknown exception occurred"
    http_code = 500

    @property
    def code(self):
        """This is here for webob to read.

        https://github.com/Pylons/webob/blob/master/webob/exc.py
        """
        return self.http_code

    def __str__(self):
        return self.message

    def __init__(self, message=None):
        if message is not None:
            self.message = message
        super(MistralException, self).__init__(
            '%d: %s' % (self.http_code, self.message))


class DataAccessException(MistralException):
    http_code = 400


class NotFoundException(MistralException):
    http_code = 404
    message = "Object not found"


class DBDuplicateEntry(MistralException):
    http_code = 409
    message = "Database object already exists"


class ActionException(MistralException):
    http_code = 400


class InvalidActionException(MistralException):
    http_code = 400


class ActionRegistrationException(MistralException):
    message = "Failed to register action"


class EngineException(MistralException):
    pass


class ApplicationContextNotFoundException(MistralException):
    http_code = 400
    message = "Application context not found"


class InvalidModelException(MistralException):
    http_code = 400
    message = "Wrong entity definition"



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo.config import cfg

from mistral.actions import std_actions
from mistral.db import api as db_api
from mistral import engine
from mistral.engine.drivers.default import engine as concrete_engine
from mistral.engine import states
from mistral import expressions
from mistral.openstack.common import log as logging
from mistral.tests import base


LOG = logging.getLogger(__name__)

WB_NAME = "my_workbook"
CONTEXT = None  # TODO(rakhmerov): Use a meaningful value.

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

# TODO(rakhmerov): add more tests for errors, execution stop etc.


@mock.patch.object(
    engine.EngineClient, 'start_workflow_execution',
    mock.MagicMock(side_effect=base.EngineTestCase.mock_start_workflow))
@mock.patch.object(
    engine.EngineClient, 'convey_task_result',
    mock.MagicMock(side_effect=base.EngineTestCase.mock_task_result))
@mock.patch.object(
    std_actions.HTTPAction, 'run',
    mock.MagicMock(return_value={'state': states.SUCCESS}))
class TestEngine(base.EngineTestCase):
    @mock.patch.object(
        concrete_engine.DefaultEngine, "_notify_task_executors",
        mock.MagicMock(return_value=""))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/one_async_task.yaml')}))
    def test_with_one_task(self):
        execution = self.engine.start_workflow_execution(WB_NAME, "create-vms",
                                                         CONTEXT)

        task = db_api.tasks_get(workbook_name=WB_NAME,
                                execution_id=execution['id'])[0]

        self.engine.convey_task_result(task['id'], states.SUCCESS, None)

        task = db_api.tasks_get(workbook_name=WB_NAME,
                                execution_id=execution['id'])[0]
        execution = db_api.execution_get(execution['id'])

        self.assertEqual(execution['state'], states.SUCCESS)
        self.assertEqual(task['state'], states.SUCCESS)

    @mock.patch.object(
        engine.EngineClient, 'get_workflow_execution_state',
        mock.MagicMock(
            side_effect=base.EngineTestCase.mock_get_workflow_state))
    @mock.patch.object(
        concrete_engine.DefaultEngine, "_notify_task_executors",
        mock.MagicMock(return_value=""))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/require_flow.yaml')}))
    def test_require_flow(self):
        execution = self.engine.start_workflow_execution(WB_NAME, "backup-vms",
                                                         CONTEXT)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self.engine.convey_task_result(tasks[0]['id'],
                                       states.SUCCESS, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self.assertIsNotNone(tasks)
        self.assertEqual(2, len(tasks))
        self.assertEqual(tasks[0]['state'], states.SUCCESS)

        # Since we mocked out executor notification we expect IDLE
        # for the second task.
        self.assertEqual(tasks[1]['state'], states.IDLE)
        self.assertEqual(states.RUNNING,
                         self.engine.get_workflow_execution_state(
                             WB_NAME, execution['id']))

        self.engine.convey_task_result(tasks[1]['id'],
                                       states.SUCCESS, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        execution = db_api.execution_get(execution['id'])

        self.assertEqual(execution['state'], states.SUCCESS)
        self.assertEqual(tasks[0]['state'], states.SUCCESS)
        self.assertEqual(tasks[1]['state'], states.SUCCESS)
        self.assertEqual(states.SUCCESS,
                         self.engine.get_workflow_execution_state(
                             WB_NAME, execution['id']))

    @mock.patch.object(
        concrete_engine.DefaultEngine, '_run_tasks',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_run_tasks))
    @mock.patch.object(
        expressions, "evaluate", mock.MagicMock(side_effect=lambda x, y: x))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/one_sync_task.yaml')}))
    def test_with_one_sync_task(self):
        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         "create-vm-nova",
                                                         CONTEXT)

        task = db_api.tasks_get(workbook_name=WB_NAME,
                                execution_id=execution['id'])[0]
        execution = db_api.execution_get(execution['id'])

        self.assertEqual(execution['state'], states.SUCCESS)
        self.assertEqual(task['state'], states.SUCCESS)

    @mock.patch.object(
        concrete_engine.DefaultEngine, '_run_tasks',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_run_tasks))
    @mock.patch.object(
        expressions, "evaluate", mock.MagicMock(side_effect=lambda x, y: x))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/direct_flow.yaml')}))
    def test_direct_flow_on_success_finish(self):
        # Start workflow.
        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         "start-task",
                                                         CONTEXT)
        # Only the first task is RUNNING
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        self.assertEqual(len(tasks), 1)
        task = self._assert_single_item(tasks,
                                        name='start-task',
                                        state=states.RUNNING)

        # Make 'start-task' successful.
        self.engine.convey_task_result(task['id'],
                                       states.SUCCESS, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        self.assertEqual(len(tasks), 3)
        self._assert_single_item(tasks,
                                 name='start-task',
                                 state=states.SUCCESS)
        task1 = self._assert_single_item(tasks,
                                         name='task-one',
                                         state=states.RUNNING)
        self._assert_single_item(tasks,
                                 name='task-two',
                                 state=states.RUNNING)

        # Make 'task-one' tasks successful.
        self.engine.convey_task_result(task1['id'],
                                       states.SUCCESS, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        tasks_2 = self._assert_multiple_items(tasks, 2,
                                              name='task-two',
                                              state=states.RUNNING)

        # Make both 'task-two' task successful.
        self.engine.convey_task_result(tasks_2[0]['id'],
                                       states.SUCCESS, None)
        self.engine.convey_task_result(tasks_2[1]['id'],
                                       states.SUCCESS, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        execution = db_api.execution_get(execution['id'])

        self._assert_multiple_items(tasks, 4, state=states.SUCCESS)
        self.assertEqual(execution['state'], states.SUCCESS)

    @mock.patch.object(
        concrete_engine.DefaultEngine, '_run_tasks',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_run_tasks))
    @mock.patch.object(
        expressions, "evaluate", mock.MagicMock(side_effect=lambda x, y: x))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/direct_flow.yaml')}))
    def test_direct_flow_on_error_finish(self):
        # Start workflow.
        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         "start-task",
                                                         CONTEXT)
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self.assertEqual(execution['state'], states.RUNNING)
        start_task = self._assert_single_item(tasks,
                                              name='start-task',
                                              state=states.RUNNING)

        # Make 'start-task' task fail.
        self.engine.convey_task_result(start_task['id'],
                                       states.ERROR, CONTEXT)
        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])

        self.assertEqual(len(tasks), 4)
        task3 = self._assert_single_item(tasks,
                                         name='task-three',
                                         state=states.RUNNING)
        task2 = self._assert_single_item(tasks,
                                         name='task-two',
                                         state=states.RUNNING)
        task4 = self._assert_single_item(tasks,
                                         name='task-four',
                                         state=states.RUNNING)

        # Make all running tasks successful.
        self.engine.convey_task_result(task2['id'],
                                       states.SUCCESS, None)
        self.engine.convey_task_result(task3['id'],
                                       states.SUCCESS, None)
        self.engine.convey_task_result(task4['id'],
                                       states.SUCCESS, None)

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        execution = db_api.execution_get(execution['id'])

        self._assert_multiple_items(tasks, 3, state=states.SUCCESS)
        self._assert_single_item(tasks, state=states.ERROR)
        self.assertEqual(execution['state'], states.SUCCESS)

    @mock.patch.object(
        concrete_engine.DefaultEngine, "_notify_task_executors",
        mock.MagicMock(return_value=""))
    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/no_namespaces.yaml')}))
    @mock.patch.object(
        concrete_engine.DefaultEngine, '_run_tasks',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_run_tasks))
    def test_engine_with_no_namespaces(self):
        execution = self.engine.start_workflow_execution(WB_NAME, "task1", {})

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        execution = db_api.execution_get(execution['id'])

        self.assertIsNotNone(tasks)
        self.assertEqual(1, len(tasks))
        self.assertEqual(tasks[0]['state'], states.SUCCESS)
        self.assertEqual(execution['state'], states.SUCCESS)

    @mock.patch.object(
        db_api, 'workbook_get',
        mock.MagicMock(return_value={'definition': base.get_resource(
            'control_flow/one_std_task.yaml')}))
    @mock.patch.object(
        concrete_engine.DefaultEngine, '_run_tasks',
        mock.MagicMock(side_effect=base.EngineTestCase.mock_run_tasks))
    def test_engine_task_std_action_with_namespaces(self):
        execution = self.engine.start_workflow_execution(WB_NAME,
                                                         "std_http_task", {})

        tasks = db_api.tasks_get(workbook_name=WB_NAME,
                                 execution_id=execution['id'])
        execution = db_api.execution_get(execution['id'])

        self.assertEqual(1, len(tasks))
        self.assertEqual(states.SUCCESS, tasks[0]['state'])
        self.assertEqual(states.SUCCESS, execution['state'])



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mistral.db import api as db_api
from mistral import exceptions
from mistral.tests.api import base

DEFINITION = "my definition"

NEW_DEFINITION = """
Namespaces:
  Service:
    actions:
      action:
        class: std.echo
        base-parameters:
            output: Haha

Workflow:
  tasks:
    task1:
      parameters:
      action: Service:action

triggers:
  create-vms:
    type: periodic
    tasks: create-vms
    parameters:
      cron-pattern: "* * * * *"
"""


class TestWorkbookDefinitionController(base.FunctionalTest):
    @mock.patch.object(db_api, "workbook_definition_get",
                       mock.MagicMock(return_value=DEFINITION))
    def test_get(self):
        resp = self.app.get('/v1/workbooks/my_workbook/definition',
                            headers={"Content-Type": "text/plain"})

        self.assertEqual(resp.status_int, 200)
        self.assertEqual(DEFINITION, resp.text)

    @mock.patch.object(db_api, "workbook_definition_get",
                       mock.MagicMock(
                           side_effect=exceptions.NotFoundException()))
    def test_get_not_found(self):
        resp = self.app.get('/v1/workbooks/my_workbook/definition',
                            headers={"Content-Type": "text/plain"},
                            expect_errors=True)

        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, "workbook_definition_put",
                       mock.MagicMock(return_value={
                           'name': 'my_workbook',
                           'definition': NEW_DEFINITION
                       }))
    def test_put(self):
        resp = self.app.put('/v1/workbooks/my_workbook/definition',
                            NEW_DEFINITION,
                            headers={"Content-Type": "text/plain"})

        self.assertEqual(resp.status_int, 200)
        self.assertEqual(NEW_DEFINITION, resp.body)

        # Check that associated triggers have been created in DB.
        triggers = db_api.triggers_get(workbook_name='my_workbook')

        self.assertEqual(triggers[0]['name'], 'create-vms')
        self.assertEqual(triggers[0]['pattern'], '* * * * *')
        self.assertEqual(triggers[0]['workbook_name'], 'my_workbook')

    @mock.patch.object(db_api, "workbook_definition_put",
                       mock.MagicMock(
                           side_effect=exceptions.NotFoundException()))
    def test_put_not_found(self):
        resp = self.app.put('/v1/workbooks/my_workbook/definition',
                            NEW_DEFINITION,
                            headers={"Content-Type": "text/plain"},
                            expect_errors=True)

        self.assertEqual(resp.status_int, 404)



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

##############################################################################
##############################################################################
##
## DO NOT MODIFY THIS FILE
##
## This file is being graduated to the oslotest library. Please make all
## changes there, and only backport critical fixes here. - dhellmann
##
##############################################################################
##############################################################################

"""Common utilities used in testing"""

import logging
import os
import tempfile

import fixtures
import testtools

_TRUE_VALUES = ('True', 'true', '1', 'yes')
_LOG_FORMAT = "%(levelname)8s [%(name)s] %(message)s"


class BaseTestCase(testtools.TestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self._set_timeout()
        self._fake_output()
        self._fake_logs()
        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())
        self.tempdirs = []

    def _set_timeout(self):
        test_timeout = os.environ.get('OS_TEST_TIMEOUT', 0)
        try:
            test_timeout = int(test_timeout)
        except ValueError:
            # If timeout value is invalid do not set a timeout.
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

    def _fake_output(self):
        if os.environ.get('OS_STDOUT_CAPTURE') in _TRUE_VALUES:
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if os.environ.get('OS_STDERR_CAPTURE') in _TRUE_VALUES:
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

    def _fake_logs(self):
        if os.environ.get('OS_DEBUG') in _TRUE_VALUES:
            level = logging.DEBUG
        else:
            level = logging.INFO
        capture_logs = os.environ.get('OS_LOG_CAPTURE') in _TRUE_VALUES
        if capture_logs:
            self.useFixture(
                fixtures.FakeLogger(
                    format=_LOG_FORMAT,
                    level=level,
                    nuke_handlers=capture_logs,
                )
            )
        else:
            logging.basicConfig(format=_LOG_FORMAT, level=level)

    def create_tempfiles(self, files, ext='.conf'):
        tempfiles = []
        for (basename, contents) in files:
            if not os.path.isabs(basename):
                (fd, path) = tempfile.mkstemp(prefix=basename, suffix=ext)
            else:
                path = basename + ext
                fd = os.open(path, os.O_CREAT | os.O_WRONLY)
            tempfiles.append(path)
            try:
                os.write(fd, contents)
            finally:
                os.close(fd)
        return tempfiles





# Copyright 2013 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import uuid

from tempest import exceptions
from tempest import test

from functionaltests.api import base


class SanityTests(base.TestCase):

    @test.attr(type='smoke')
    def test_get_list_obj(self):
        resp, _ = self.client.get_list_obj('')
        self.assertEqual(200, resp.status)

    @test.attr(type='smoke')
    def test_get_list_workbooks(self):
        resp, body = self.client.get_list_obj('workbooks')

        self.assertEqual(200, resp.status)
        self.assertEqual([], body['workbooks'])

    @test.attr(type='smoke')
    def test_get_workbook(self):
        self.client.create_obj('workbooks', 'test')
        self.obj.append(['workbooks', 'test'])
        resp, body = self.client.get_list_obj('workbooks/test')

        self.assertEqual(200, resp.status)
        self.assertEqual('test', body['name'])

    @test.attr(type='smoke')
    def test_get_executions(self):
        self.client.create_obj('workbooks', 'test')
        self.obj.append(['workbooks', 'test'])
        resp, body = self.client.get_list_obj('workbooks/test/executions')

        self.assertEqual(200, resp.status)
        self.assertEqual([], body['executions'])

    @test.attr(type='smoke')
    def test_create_and_delete_workbook(self):
        resp, body = self.client.create_obj('workbooks', 'test')
        self.obj.append(['workbooks', 'test'])

        self.assertEqual(201, resp.status)
        self.assertEqual('test', body['name'])

        resp, body = self.client.get_list_obj('workbooks')

        self.assertEqual(200, resp.status)
        self.assertEqual('test', body['workbooks'][0]['name'])

        self.client.delete_obj('workbooks', 'test')
        _, body = self.client.get_list_obj('workbooks')

        self.assertEqual([], body['workbooks'])

    @test.attr(type='smoke')
    def test_update_workbook(self):
        self.client.create_obj('workbooks', 'test')
        self.obj.append(['workbooks', 'test'])
        resp, body = self.client.update_obj('workbooks', 'test')

        self.assertEqual(200, resp.status)
        self.assertEqual('testupdated', body['name'])

        self.obj.append(['workbooks', 'testupdated'])

    @test.attr(type='smoke')
    def test_get_workbook_definition(self):
        self.client.create_obj('workbooks', 'test')
        self.obj.append(['workbooks', 'test'])
        self.client.upload_workbook_definition('test')
        resp, body = self.client.get_workbook_definition('test')

        self.assertEqual(200, resp.status)
        self.assertIsNotNone(body)

    @test.attr(type='smoke')
    def test_upload_workbook_definition(self):
        self.client.create_obj('workbooks', 'test1')
        self.obj.append(['workbooks', 'test1'])
        resp, body = self.client.upload_workbook_definition('test1')

        self.assertEqual(200, resp.status)
        self.assertIsNotNone(body)

    @test.attr(type='negative')
    def test_double_create_obj(self):
        self.client.create_obj('workbooks', 'test')
        self.obj.append(['workbooks', 'test'])

        self.assertRaises(exceptions.Conflict, self.client.create_obj,
                          'workbooks', 'test')

        self.client.delete_obj('workbooks', 'test')
        _, body = self.client.get_list_obj('workbooks')

        self.assertEqual([], body['workbooks'])

    @test.attr(type='negative')
    def test_get_nonexistent_workbook_definition(self):
        self.assertRaises(exceptions.NotFound,
                          self.client.get_workbook_definition,
                          'fksn')

    @test.attr(type='negative')
    def test_get_executions_list_from_nonexistent_workbook(self):
        self.assertRaises(exceptions.NotFound, self.client.get_list_obj,
                          'workbooks/nonexistentworkbook/executions')


class AdvancedTests(base.TestCaseAdvanced):

    @test.attr(type='positive')
    def test_create_execution(self):
        resp, body = self._create_execution('test123', 'aloha')

        self.assertEqual(201, resp.status)
        self.assertEqual('test123', body["workbook_name"])

    @test.attr(type='positive')
    def test_get_execution(self):
        _, execution = self._create_execution('test123', 'aloha')

        resp, body = self.client.get_execution('test123', execution['id'])

        body = json.loads(body)
        del execution['state']
        del body['state']

        self.assertEqual(200, resp.status)
        self.assertEqual(execution, body)

    # TODO(smurashov): Need to add test which would check execution update

    @test.attr(type='positive')
    def test_get_tasks_list(self):
        execution = self._create_execution('test123', 'aloha')[1]

        resp, tasks = self.client.get_tasks_list('test123', execution['id'])

        self.assertEqual(200, resp.status)
        for task in tasks:
            self.assertEqual(execution['id'], task['execution_id'])
            self.assertEqual('test123', task['workbook_name'])

    @test.attr(type='positive')
    def test_get_task(self):
        _, execution = self._create_execution('test123', 'aloha')

        tasks = self.client.get_tasks_list('test123', execution['id'])[1]

        resp, task = self.client.get_task('test123', execution['id'],
                                          tasks[0]['id'])

        del tasks[0]['state']
        del task['state']

        self.assertEqual(200, resp.status)
        self.assertEqual(tasks[0], task)

    # TODO(smurashov): Need to add test which would check task update

    @test.attr(type='negative')
    def test_create_execution_in_nonexistent_workbook(self):
        self.assertRaises(exceptions.NotFound, self._create_execution,
                          'nonexistentworkbook', 'catchme')

    def test_get_nonexistent_execution(self):

        self.assertRaises(exceptions.NotFound, self.client.get_execution,
                          'test123', str(uuid.uuid4()))

    @test.attr(type='negative')
    def test_update_nonexistent_execution(self):

        id = str(uuid.uuid4())
        put_body = {
            "id": id,
            "state": "STOPPED"
        }

        self.assertRaises(exceptions.NotFound, self.client.update_execution,
                          'test123', id, put_body)

    @test.attr(type='negative')
    def test_get_tasks_list_of_nonexistent_execution(self):

        self.assertRaises(exceptions.NotFound, self.client.get_tasks_list,
                          'test123', str(uuid.uuid4()))


class AdvancedNegativeTestsWithExecutionCreate(base.TestCaseAdvanced):

    def setUp(self):
        super(AdvancedNegativeTestsWithExecutionCreate, self).setUp()

        self.execution = self._create_execution('test123', 'aloha')[1]

    @test.attr(type='negative')
    def test_get_nonexistent_task(self):
        self.assertRaises(exceptions.NotFound, self.client.get_task,
                          'test123', self.execution['id'], str(uuid.uuid4()))

    @test.attr(type='negative')
    def test_update_nonexistent_task(self):

        id = str(uuid.uuid4())
        put_body = {
            "id": id,
            "state": "ERROR"
        }

        self.assertRaises(exceptions.NotFound, self.client.update_task,
                          'test123', self.execution['id'], str(uuid.uuid4()),
                          put_body)



# -*- coding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
import copy

import eventlet
from oslo.config import cfg
from oslo import messaging
import six
from stevedore import driver

# Submoules of mistral.engine will throw NoSuchOptError if configuration
# options required at top level of this  __init__.py are not imported before
# the submodules are referenced.
cfg.CONF.import_opt('workflow_trace_log_name', 'mistral.config')

from mistral import context as auth_context
from mistral.db import api as db_api
from mistral import dsl_parser as parser
from mistral.engine import data_flow
from mistral.engine import retry
from mistral.engine import states
from mistral.engine import workflow
from mistral import exceptions as exc
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)
WORKFLOW_TRACE = logging.getLogger(cfg.CONF.workflow_trace_log_name)


def get_transport(transport=None):
    return (transport if transport else messaging.get_transport(cfg.CONF))


def get_engine(name, transport):
    mgr = driver.DriverManager(
        namespace='mistral.engine.drivers',
        name=name,
        invoke_on_load=True,
        invoke_kwds={'transport': transport})
    return mgr.driver


@six.add_metaclass(abc.ABCMeta)
class Engine(object):
    """Abstract engine for workflow execution."""

    transport = None

    def __init__(self, transport=None):
        self.transport = get_transport(transport)

    @abc.abstractmethod
    def _run_tasks(cls, tasks):
        raise NotImplementedError()

    def start_workflow_execution(self, cntx, **kwargs):
        """Starts a workflow execution based on the specified workbook name
        and target task.

        :param cntx: a request context dict
        :type cntx: MistralContext
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        :return: Workflow execution.
        """
        workbook_name = kwargs.get('workbook_name')
        task_name = kwargs.get('task_name')
        context = kwargs.get('context', None)

        context = copy.copy(context) if context else {}

        WORKFLOW_TRACE.info("New execution started - [workbook_name = '%s', "
                            "task_name = '%s']" % (workbook_name, task_name))

        db_api.start_tx()

        # Persist execution and tasks in DB.
        try:
            workbook = self._get_workbook(workbook_name)
            execution = self._create_execution(workbook_name,
                                               task_name,
                                               context)

            tasks = self._create_tasks(
                workflow.find_workflow_tasks(workbook, task_name),
                workbook,
                workbook_name, execution['id']
            )

            tasks_to_start, delayed_tasks = workflow.find_resolved_tasks(tasks)

            self._add_variables_to_data_flow_context(context, execution)

            data_flow.prepare_tasks(tasks_to_start, context)

            db_api.commit_tx()
        except Exception as e:
            msg = "Failed to create necessary DB objects: %s" % e
            LOG.exception(msg)
            raise exc.EngineException(msg)
        finally:
            db_api.end_tx()

        for task in delayed_tasks:
            self._schedule_run(workbook, task, context)

        self._run_tasks(tasks_to_start)

        return execution

    def stop_workflow_execution(self, cntx, **kwargs):
        """Stops the workflow execution with the given id.

        :param cntx: a request context dict
        :type cntx: dict
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        :return: Workflow execution.
        """
        execution_id = kwargs.get('execution_id')

        return db_api.execution_update(execution_id,
                                       {"state": states.STOPPED})

    def convey_task_result(self, cntx, **kwargs):
        """Conveys task result to Mistral Engine.

        This method should be used by clients of Mistral Engine to update
        state of a task once task action has been performed. One of the
        clients of this method is Mistral REST API server that receives
        task result from the outside action handlers.

        Note: calling this method serves an event notifying Mistral that
        it possibly needs to move the workflow on, i.e. run other workflow
        tasks for which all dependencies are satisfied.

        :param cntx: a request context dict
        :type cntx: dict
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        :return: Task.
        """
        task_id = kwargs.get('task_id')
        state = kwargs.get('state')
        result = kwargs.get('result')

        db_api.start_tx()

        try:
            # TODO(rakhmerov): validate state transition
            task = db_api.task_get(task_id)
            workbook = self._get_workbook(task['workbook_name'])

            wf_trace_msg = "Task '%s' [%s -> %s" % \
                           (task['name'], task['state'], state)

            wf_trace_msg += ']' if state == states.ERROR \
                else ", result = %s]" % result
            WORKFLOW_TRACE.info(wf_trace_msg)

            task_output = data_flow.get_task_output(task, result)

            # Update task state.
            task, outbound_context = self._update_task(workbook, task, state,
                                                       task_output)

            execution = db_api.execution_get(task['execution_id'])

            self._create_next_tasks(task, workbook)

            # Determine what tasks need to be started.
            tasks = db_api.tasks_get(workbook_name=task['workbook_name'],
                                     execution_id=task['execution_id'])

            new_exec_state = self._determine_execution_state(execution, tasks)

            if execution['state'] != new_exec_state:
                wf_trace_msg = \
                    "Execution '%s' [%s -> %s]" % \
                    (execution['id'], execution['state'], new_exec_state)
                WORKFLOW_TRACE.info(wf_trace_msg)

                execution = \
                    db_api.execution_update(execution['id'], {
                        "state": new_exec_state
                    })

                LOG.info("Changed execution state: %s" % execution)

            tasks_to_start, delayed_tasks = workflow.find_resolved_tasks(tasks)

            self._add_variables_to_data_flow_context(outbound_context,
                                                     execution)

            data_flow.prepare_tasks(tasks_to_start, outbound_context)

            db_api.commit_tx()
        except Exception as e:
            msg = "Failed to create necessary DB objects: %s" % e
            LOG.exception(msg)
            raise exc.EngineException(msg)
        finally:
            db_api.end_tx()

        if states.is_stopped_or_finished(execution["state"]):
            return task

        for task in delayed_tasks:
            self._schedule_run(workbook, task, outbound_context)

        if tasks_to_start:
            self._run_tasks(tasks_to_start)

        return task

    def get_workflow_execution_state(self, cntx, **kwargs):
        """Gets the workflow execution state.

        :param cntx: a request context dict
        :type cntx: dict
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        :return: Current workflow state.
        """
        workbook_name = kwargs.get('workbook_name')
        execution_id = kwargs.get('execution_id')

        execution = db_api.execution_get(execution_id)

        if not execution:
            raise exc.EngineException("Workflow execution not found "
                                      "[workbook_name=%s, execution_id=%s]"
                                      % (workbook_name, execution_id))

        return execution["state"]

    def get_task_state(self, cntx, **kwargs):
        """Gets task state.

        :param cntx: a request context dict
        :type cntx: dict
        :param kwargs: a dict of method arguments
        :type kwargs: dict
        :return: Current task state.
        """
        task_id = kwargs.get('task_id')

        task = db_api.task_get(task_id)

        if not task:
            raise exc.EngineException("Task not found.")

        return task["state"]

    @classmethod
    def _create_execution(cls, workbook_name, task_name, context):
        return db_api.execution_create(workbook_name, {
            "workbook_name": workbook_name,
            "task": task_name,
            "state": states.RUNNING,
            "context": context
        })

    @classmethod
    def _add_variables_to_data_flow_context(cls, df_ctx, execution):
        db_workbook = db_api.workbook_get(execution['workbook_name'])

        data_flow.add_openstack_data_to_context(df_ctx, db_workbook)
        data_flow.add_execution_to_context(df_ctx, execution)

    @classmethod
    def _create_next_tasks(cls, task, workbook):
        tasks = workflow.find_tasks_after_completion(task, workbook)

        db_tasks = cls._create_tasks(tasks, workbook, task['workbook_name'],
                                     task['execution_id'])
        return workflow.find_resolved_tasks(db_tasks)

    @classmethod
    def _create_tasks(cls, task_list, workbook, workbook_name, execution_id):
        tasks = []

        for task in task_list:
            state, task_runtime_context = retry.get_task_runtime(task)
            action_spec = workbook.get_action(task.get_full_action_name())

            db_task = db_api.task_create(execution_id, {
                "name": task.name,
                "requires": task.get_requires(),
                "task_spec": task.to_dict(),
                "action_spec": {} if not action_spec
                else action_spec.to_dict(),
                "state": state,
                "tags": task.get_property("tags", None),
                "task_runtime_context": task_runtime_context,
                "workbook_name": workbook_name
            })

            tasks.append(db_task)

        return tasks

    @classmethod
    def _get_workbook(cls, workbook_name):
        wb = db_api.workbook_get(workbook_name)
        return parser.get_workbook(wb["definition"])

    @classmethod
    def _determine_execution_state(cls, execution, tasks):
        if workflow.is_error(tasks):
            return states.ERROR

        if workflow.is_success(tasks) or workflow.is_finished(tasks):
            return states.SUCCESS

        return execution['state']

    @classmethod
    def _update_task(cls, workbook, task, state, task_output):
        """Update the task with the runtime information. The outbound_context
        for this task is also calculated.
        :return: task, outbound_context. task is the updated task and
        computed outbound context.
        """
        task_spec = workbook.tasks.get(task["name"])
        task_runtime_context = task["task_runtime_context"]

        # Compute the outbound_context, state and exec_flow_context.
        outbound_context = data_flow.get_outbound_context(task, task_output)
        state, task_runtime_context = retry.get_task_runtime(
            task_spec, state, outbound_context, task_runtime_context)

        # Update the task.
        update_values = {"state": state,
                         "output": task_output,
                         "task_runtime_context": task_runtime_context}
        task = db_api.task_update(task["id"], update_values)

        return task, outbound_context

    def _schedule_run(cls, workbook, task, outbound_context):
        """Schedules task to run after the delay defined in the task
        specification. If no delay is specified this method is a no-op.
        """

        # TODO(rakhmerov): Reavaluate parameter 'context' once it's clear
        # how to work with trust chains correctly in keystone
        # (waiting for corresponding changes to be made).
        def run_delayed_task(context):
            """Runs the delayed task. Performs all the steps required to setup
            a task to run which are not already done. This is mostly code
            copied over from convey_task_result.

            :param context Mistral authentication context inherited from a
                caller thread.
            """
            auth_context.set_ctx(context)

            db_api.start_tx()

            try:
                execution_id = task['execution_id']
                execution = db_api.execution_get(execution_id)

                # Change state from DELAYED to IDLE to unblock processing.

                WORKFLOW_TRACE.info("Task '%s' [%s -> %s]"
                                    % (task['name'],
                                       task['state'], states.IDLE))

                db_task = db_api.task_update(task['id'],
                                             {"state": states.IDLE})
                task_to_start = [db_task]
                data_flow.prepare_tasks(task_to_start, outbound_context)
                db_api.commit_tx()
            finally:
                db_api.end_tx()

            if not states.is_stopped_or_finished(execution["state"]):
                cls._run_tasks(task_to_start)

        task_spec = workbook.tasks.get(task['name'])
        retries, break_on, delay_sec = task_spec.get_retry_parameters()

        if delay_sec > 0:
            # Run the task after the specified delay.
            eventlet.spawn_after(delay_sec, run_delayed_task,
                                 context=auth_context.ctx())
        else:
            LOG.warn("No delay specified for task(id=%s) name=%s. Not "
                     "scheduling for execution." % (task['id'], task['name']))


class EngineClient(object):
    """RPC client for the Engine."""

    def __init__(self, transport):
        """Construct an RPC client for the Engine.

        :param transport: a messaging transport handle
        :type transport: Transport
        """
        serializer = auth_context.RpcContextSerializer(
            auth_context.JsonPayloadSerializer())
        target = messaging.Target(topic=cfg.CONF.engine.topic)
        self._client = messaging.RPCClient(transport, target,
                                           serializer=serializer)

    def start_workflow_execution(self, workbook_name, task_name, context=None):
        """Starts a workflow execution based on the specified workbook name
        and target task.

        :param workbook_name: Workbook name
        :param task_name: Target task name
        :param context: Execution context which defines a workflow input
        :return: Workflow execution.
        """
        cntx = auth_context.ctx()
        kwargs = {'workbook_name': workbook_name,
                  'task_name': task_name,
                  'context': context}
        return self._client.call(cntx, 'start_workflow_execution', **kwargs)

    def stop_workflow_execution(self, workbook_name, execution_id):
        """Stops the workflow execution with the given id.

        :param workbook_name: Workbook name.
        :param execution_id: Workflow execution id.
        :return: Workflow execution.
        """
        cntx = auth_context.ctx()
        kwargs = {'workbook_name': workbook_name,
                  'execution_id': execution_id}
        return self._client.call(cntx, 'stop_workflow_execution', **kwargs)

    def convey_task_result(self, task_id, state, result):
        """Conveys task result to Mistral Engine.

        This method should be used by clients of Mistral Engine to update
        state of a task once task action has been performed. One of the
        clients of this method is Mistral REST API server that receives
        task result from the outside action handlers.

        Note: calling this method serves an event notifying Mistral that
        it possibly needs to move the workflow on, i.e. run other workflow
        tasks for which all dependencies are satisfied.

        :param task_id: Task id.
        :param state: New task state.
        :param result: Task result data.
        :return: Task.
        """
        cntx = auth_context.ctx()
        kwargs = {'task_id': task_id,
                  'state': state,
                  'result': result}
        return self._client.call(cntx, 'convey_task_result', **kwargs)

    def get_workflow_execution_state(self, workbook_name, execution_id):
        """Gets the workflow execution state.

        :param workbook_name: Workbook name.
        :param execution_id: Workflow execution id.
        :return: Current workflow state.
        """
        cntx = auth_context.ctx()
        kwargs = {'workbook_name': workbook_name,
                  'execution_id': execution_id}
        return self._client.call(
            cntx, 'get_workflow_execution_state', **kwargs)

    def get_task_state(self, workbook_name, execution_id, task_id):
        """Gets task state.

        :param workbook_name: Workbook name.
        :param execution_id: Workflow execution id.
        :param task_id: Task id.
        :return: Current task state.
        """
        cntx = auth_context.ctx()
        kwargs = {'workbook_name': workbook_name,
                  'executioin_id': execution_id,
                  'task_id': task_id}
        return self._client.call(cntx, 'get_task_state', **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import expressions as expr
from mistral.tests import base

DATA = {
    "server": {
        "id": "03ea824a-aa24-4105-9131-66c48ae54acf",
        "name": "cloud-fedora",
        "status": "ACTIVE"
    },
    "status": "OK"
}

SERVERS = {
    "servers": [
        {
            'name': 'centos'
        },
        {
            'name': 'ubuntu'
        },
        {
            'name': 'fedora'
        }
    ]
}


class YaqlEvaluatorTest(base.BaseTest):
    def setUp(self):
        super(YaqlEvaluatorTest, self).setUp()
        self._evaluator = expr.YAQLEvaluator()

    def test_expression_result(self):
        res = self._evaluator.evaluate('$.server', DATA)
        self.assertEqual(res, {
            'id': "03ea824a-aa24-4105-9131-66c48ae54acf",
            'name': 'cloud-fedora',
            'status': 'ACTIVE'
        })

        res = self._evaluator.evaluate('$.server.id', DATA)
        self.assertEqual(res, '03ea824a-aa24-4105-9131-66c48ae54acf')

        res = self._evaluator.evaluate("$.server.status = 'ACTIVE'", DATA)
        self.assertTrue(res)

    def test_wrong_expression(self):
        res = self._evaluator.evaluate("$.status = 'Invalid value'", DATA)
        self.assertFalse(res)

        res = self._evaluator.evaluate('$.wrong_key', DATA)
        self.assertIsNone(res)

        expression_str = 'invalid_expression_string'
        res = self._evaluator.evaluate(expression_str, DATA)
        self.assertEqual(res, expression_str)

    def test_select_result(self):
        res = self._evaluator.evaluate('$.servers[$.name = ubuntu]', SERVERS)
        item = list(res)[0]
        self.assertEqual(item, {'name': 'ubuntu'})

    def test_evaluate_recursively(self):
        task_spec_dict = {
            'parameters': {
                'p1': 'My string',
                'p2': '$.param2',
                'p3': ''
            },
            'publish': {
                'new_key11': 'new_key1'
            }
        }
        modified_task = expr.evaluate_recursively(task_spec_dict,
                                                  {
                                                      'param2': 'val32'
                                                  })

        self.assertDictEqual(
            {
                'parameters': {
                    'p1': 'My string',
                    'p2': 'val32',
                    'p3': ''
                },
                'publish': {
                    'new_key11': 'new_key1'
                }
            },
            modified_task)

    def test_evaluate_recursively_arbitrary_dict(self):
        context = {
            "auth_token": "123",
            "project_id": "mistral"
        }
        data = {
            "parameters": {
                "parameter1": {
                    "name1": "$.auth_token",
                    "name2": "val_name2"
                },
                "param2": [
                    "var1",
                    "var2",
                    "/servers/{$.project_id}/bla"
                ]
            },
            "token": "$.auth_token"
        }

        applied = expr.evaluate_recursively(data, context)

        self.assertDictEqual(
            {
                "parameters": {
                    "parameter1": {
                        "name1": "123",
                        "name2": "val_name2"
                    },
                    "param2": ["var1", "var2", "/servers/mistral/bla"]
                },
                "token": "123"
            },
            applied)



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mistral.db import api as db_api
from mistral import exceptions
from mistral.tests.api import base

WORKBOOKS = [
    {
        'name': "my_workbook",
        'description': "My cool Mistral workbook",
        'tags': ['deployment', 'demo']
    }
]

UPDATED_WORKBOOK = WORKBOOKS[0].copy()
UPDATED_WORKBOOK['description'] = 'new description'


class TestWorkbooksController(base.FunctionalTest):
    @mock.patch.object(db_api, "workbook_get",
                       mock.MagicMock(return_value=WORKBOOKS[0]))
    def test_get(self):
        resp = self.app.get('/v1/workbooks/my_workbook')

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(WORKBOOKS[0], resp.json)

    @mock.patch.object(db_api, "workbook_get",
                       mock.MagicMock(
                           side_effect=exceptions.NotFoundException()))
    def test_get_not_found(self):
        resp = self.app.get('/v1/workbooks/dev_null', expect_errors=True)
        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, "workbook_update",
                       mock.MagicMock(return_value=UPDATED_WORKBOOK))
    def test_put(self):
        resp = self.app.put_json('/v1/workbooks/my_workbook',
                                 dict(description='new description'))

        self.assertEqual(resp.status_int, 200)
        self.assertDictEqual(UPDATED_WORKBOOK, resp.json)

    @mock.patch.object(db_api, "workbook_update",
                       mock.MagicMock(
                           side_effect=exceptions.NotFoundException()))
    def test_put_not_found(self):
        resp = self.app.put_json('/v1/workbooks/my_workbook',
                                 dict(description='new description'),
                                 expect_errors=True)
        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, "workbook_create",
                       mock.MagicMock(return_value=WORKBOOKS[0]))
    @mock.patch("mistral.services.trusts.create_trust",
                mock.MagicMock(return_value=WORKBOOKS[0]))
    def test_post(self):
        resp = self.app.post_json('/v1/workbooks', WORKBOOKS[0])

        self.assertEqual(resp.status_int, 201)
        self.assertDictEqual(WORKBOOKS[0], resp.json)

    @mock.patch.object(db_api, "workbook_create",
                       mock.MagicMock(side_effect=exceptions.DBDuplicateEntry))
    @mock.patch("mistral.services.trusts.create_trust",
                mock.MagicMock(return_value=WORKBOOKS[0]))
    def test_post_dup(self):
        resp = self.app.post_json('/v1/workbooks', WORKBOOKS[0],
                                  expect_errors=True)

        self.assertEqual(resp.status_int, 409)

    @mock.patch.object(db_api, "workbook_delete",
                       mock.MagicMock(return_value=None))
    def test_delete(self):
        resp = self.app.delete('/v1/workbooks/my_workbook')

        self.assertEqual(resp.status_int, 204)

    @mock.patch.object(db_api, "workbook_delete",
                       mock.MagicMock(
                           side_effect=exceptions.NotFoundException()))
    def test_delete_not_found(self):
        resp = self.app.delete('/v1/workbooks/my_workbook', expect_errors=True)

        self.assertEqual(resp.status_int, 404)

    @mock.patch.object(db_api, "workbooks_get",
                       mock.MagicMock(return_value=WORKBOOKS))
    def test_get_all(self):
        resp = self.app.get('/v1/workbooks')

        self.assertEqual(resp.status_int, 200)

        self.assertEqual(len(resp.json), 1)
        self.assertDictEqual(WORKBOOKS[0], resp.json['workbooks'][0])



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.workbook import actions
from mistral.workbook import base


def merge_dicts(left, right):
    for k, v in right.iteritems():
        if k not in left:
            left[k] = v
        else:
            left_v = left[k]

            if isinstance(left_v, dict) and isinstance(v, dict):
                merge_dicts(left_v, v)


def merge_base_parameters(action, ns_base_parameters):
    if not ns_base_parameters:
        return

    if 'base-parameters' not in action:
        action['base-parameters'] = ns_base_parameters
        return

    action_base_parameters = action['base-parameters']

    merge_dicts(action_base_parameters, ns_base_parameters)


class NamespaceSpec(base.BaseSpec):
    _required_keys = ['name', 'actions']

    def __init__(self, namespace):
        super(NamespaceSpec, self).__init__(namespace)

        if self.validate():
            self.name = namespace['name']
            self.clazz = namespace.get('class')
            self.base_parameters = namespace.get('base-parameters')
            self.parameters = namespace.get('parameters')

            for _, action in namespace['actions'].iteritems():
                action['namespace'] = self.name

                if 'class' not in action:
                    action['class'] = self.clazz

                merge_base_parameters(action, self.base_parameters)

            self.actions = actions.ActionSpecList(namespace['actions'])


class NamespaceSpecList(base.BaseSpecList):
    item_class = NamespaceSpec



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import networkx as nx
from networkx.algorithms import traversal

from mistral.engine import states
from mistral.openstack.common import log as logging


LOG = logging.getLogger(__name__)


def find_workflow_tasks(workbook, task_name):
    wb_tasks = workbook.tasks
    full_graph = nx.DiGraph()
    for t in wb_tasks:
        full_graph.add_node(t)

    _update_dependencies(wb_tasks, full_graph)

    graph = _get_subgraph(full_graph, task_name)
    tasks = []
    for node in graph:
        tasks.append(wb_tasks[node])

    return tasks


def find_resolved_tasks(tasks):
    # We need to analyse graph and see which tasks are ready to start
    resolved_tasks = []
    delayed_tasks = []
    allows = []
    for t in tasks:
        if t['state'] == states.SUCCESS:
            allows += [t['name']]
    allow_set = set(allows)
    for t in tasks:
        deps = t.get('requires', [])
        if len(set(deps) - allow_set) == 0:
        # all required tasks, if any, are SUCCESS
            if t['state'] == states.IDLE:
                resolved_tasks.append(t)
            elif t['state'] == states.DELAYED:
                delayed_tasks.append(t)
    return resolved_tasks, delayed_tasks


def _get_checked_tasks(target_tasks):
    checked_tasks = []
    for t in target_tasks:
        # TODO(nmakhotkin): see and evaluate YAQL with data from context
        checked_tasks.append(t)
    return checked_tasks


def _get_tasks_to_schedule(target_tasks, workbook):
    tasks_to_schedule = _get_checked_tasks(target_tasks)
    return [workbook.tasks.get(t_name) for t_name in tasks_to_schedule]


def find_tasks_after_completion(task, workbook):
    """Determine tasks which should be scheduled after completing
    given task. Expression 'on_finish' is not mutually exclusive to
    'on_success' and 'on_error'.

    :param task: Task object
    :param workbook: Workbook Entity
    :return: list of task dictionaries.
    """
    state = task['state']
    found_tasks = []
    LOG.debug("Recieved task %s: %s" % (task['name'], state))

    if state == states.ERROR:
        tasks_on_error = workbook.tasks.get(task['name']).get_on_error()
        if tasks_on_error:
            found_tasks = _get_tasks_to_schedule(tasks_on_error, workbook)

    elif state == states.SUCCESS:
        tasks_on_success = workbook.tasks.get(task['name']).get_on_success()
        if tasks_on_success:
            found_tasks = _get_tasks_to_schedule(tasks_on_success, workbook)

    if states.is_finished(state):
        tasks_on_finish = workbook.tasks.get(task['name']).get_on_finish()
        if tasks_on_finish:
            found_tasks += _get_tasks_to_schedule(tasks_on_finish, workbook)

    LOG.debug("Found tasks: %s" % found_tasks)

    workflow_tasks = []
    for t in found_tasks:
        workflow_tasks += find_workflow_tasks(workbook, t.name)

    LOG.debug("Workflow tasks to schedule: %s" % workflow_tasks)

    return workflow_tasks


def is_finished(tasks):
    return all(states.is_finished(task['state']) for task in tasks)


def is_success(tasks):
    return all(task['state'] == states.SUCCESS for task in tasks)


def is_error(tasks):
    return all(task['state'] == states.ERROR for task in tasks)


def _get_subgraph(full_graph, task_name):
    nodes_set = traversal.dfs_predecessors(full_graph.reverse(),
                                           task_name).keys()
    nodes_set.append(task_name)

    return full_graph.subgraph(nodes_set)


def _get_dependency_tasks(tasks, task):
    if len(tasks[task].requires) < 1:
        return []

    deps = set()
    for t in tasks:
        for dep in tasks[task].requires:
            if dep == t:
                deps.add(t)

    return deps


def _update_dependencies(tasks, graph):
    for task in tasks:
        for dep in _get_dependency_tasks(tasks, task):
            graph.add_edge(dep, task)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Import related utilities and helper functions.
"""

import sys
import traceback


def import_class(import_str):
    """Returns a class from a string including module and class."""
    mod_str, _sep, class_str = import_str.rpartition('.')
    __import__(mod_str)
    try:
        return getattr(sys.modules[mod_str], class_str)
    except AttributeError:
        raise ImportError('Class %s cannot be found (%s)' %
                          (class_str,
                           traceback.format_exception(*sys.exc_info())))


def import_object(import_str, *args, **kwargs):
    """Import a class and return an instance of it."""
    return import_class(import_str)(*args, **kwargs)


def import_object_ns(name_space, import_str, *args, **kwargs):
    """Tries to import object from default namespace.

    Imports a class and return an instance of it, first by trying
    to find the class in a default namespace, then failing back to
    a full path if not found in the default namespace.
    """
    import_value = "%s.%s" % (name_space, import_str)
    try:
        return import_class(import_value)(*args, **kwargs)
    except ImportError:
        return import_class(import_str)(*args, **kwargs)


def import_module(import_str):
    """Import a module."""
    __import__(import_str)
    return sys.modules[import_str]


def import_versioned_module(version, submodule=None):
    module = 'mistral.v%s' % version
    if submodule:
        module = '.'.join((module, submodule))
    return import_module(module)


def try_import(import_str, default=None):
    """Try to import a module and if it fails return default."""
    try:
        return import_module(import_str)
    except ImportError:
        return default



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers import resource
from mistral.api.controllers.v1 import task
from mistral.db import api as db_api
from mistral.openstack.common import log as logging
from mistral.utils import rest_utils


LOG = logging.getLogger(__name__)


class Execution(resource.Resource):
    """Execution resource."""

    id = wtypes.text
    workbook_name = wtypes.text
    task = wtypes.text
    state = wtypes.text
    # Context is a JSON object but since WSME doesn't support arbitrary
    # dictionaries we have to use text type convert to json and back manually.
    context = wtypes.text

    def to_dict(self):
        d = super(Execution, self).to_dict()

        if d.get('context'):
            d['context'] = json.loads(d['context'])

        return d

    @classmethod
    def from_dict(cls, d):
        e = cls()

        for key, val in d.items():
            if hasattr(e, key):
                # Nonetype check for dictionary must be explicit
                if key == 'context' and val is not None:
                    val = json.dumps(val)
                setattr(e, key, val)

        return e

    @classmethod
    def sample(cls):
        return cls(id='1234',
                   workbook_name='flow',
                   task='doit',
                   state='SUCCESS',
                   context='{}')


class Executions(resource.Resource):
    """A collection of Execution resources."""

    executions = [Execution]

    @classmethod
    def sample(cls):
        return cls(executions=[Execution.sample()])


class ExecutionsController(rest.RestController):

    def _get(self, id):
        values = db_api.execution_get(id)
        return Execution.from_dict(values)

    def _put(self, id, execution):
        values = db_api.execution_update(id, execution.to_dict())

        return Execution.from_dict(values)

    def _delete(self, id):
        db_api.execution_delete(id)

    def _get_all(self, **kwargs):
        executions = [Execution.from_dict(values) for values
                      in db_api.executions_get(**kwargs)]

        return Executions(executions=executions)


class WorkbookExecutionsController(ExecutionsController):

    tasks = task.WorkbookTasksController()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Execution, wtypes.text, wtypes.text)
    def get(self, workbook_name, id):
        """Return the specified Execution."""
        LOG.debug("Fetch execution [workbook_name=%s, id=%s]" %
                  (workbook_name, id))
        return self._get(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Execution, wtypes.text, wtypes.text, body=Execution)
    def put(self, workbook_name, id, execution):
        """Update the specified Execution."""
        LOG.debug("Update execution [workbook_name=%s, id=%s, execution=%s]" %
                  (workbook_name, id, execution))
        return self._put(id, execution)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Execution, wtypes.text, body=Execution,
                         status_code=201)
    def post(self, workbook_name, execution):
        """Create a new Execution."""
        LOG.debug("Create execution [workbook_name=%s, execution=%s]" %
                  (workbook_name, execution))

        if (db_api.workbook_get(workbook_name)
                and db_api.workbook_definition_get(workbook_name)):
            context = None
            if execution.context:
                context = json.loads(execution.context)

            engine = pecan.request.context['engine']
            values = engine.start_workflow_execution(execution.workbook_name,
                                                     execution.task,
                                                     context)

            return Execution.from_dict(values)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, wtypes.text, status_code=204)
    def delete(self, workbook_name, id):
        """Delete the specified Execution."""
        LOG.debug("Delete execution [workbook_name=%s, id=%s]" %
                  (workbook_name, id))
        return self._delete(id)

    @wsme_pecan.wsexpose(Executions, wtypes.text)
    def get_all(self, workbook_name):
        """Return all Executions."""
        LOG.debug("Fetch executions [workbook_name=%s]" % workbook_name)

        if db_api.workbook_get(workbook_name):
            return self._get_all(workbook_name=workbook_name)


class RootExecutionsController(ExecutionsController):

    tasks = task.ExecutionTasksController()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Execution, wtypes.text)
    def get(self, id):
        """Return the specified Execution."""
        LOG.debug("Fetch execution [id=%s]" % id)
        return self._get(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(Execution, wtypes.text, body=Execution)
    def put(self, id, execution):
        """Update the specified Execution."""
        LOG.debug("Update execution [id=%s, execution=%s]" %
                  (id, execution))
        return self._put(id, execution)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, id):
        """Delete the specified Execution."""
        LOG.debug("Delete execution [id=%s]" % id)
        return self._delete(id)

    @wsme_pecan.wsexpose(Executions)
    def get_all(self):
        """Return all Executions."""
        LOG.debug("Fetch executions")

        return self._get_all()



# -*- coding: utf-8 -*-
#
# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import eventlet
from oslo import messaging
from pecan import hooks

from mistral import exceptions as exc
from mistral.openstack.common import jsonutils
from mistral.openstack.common import log as logging
from mistral import utils

LOG = logging.getLogger(__name__)

# TODO(rakhmerov): refactor all threadlocal things with mistral.utils

_CTX_THREAD_LOCAL_NAME = "MISTRAL_APP_CTX_THREAD_LOCAL"


class BaseContext(object):
    """Container for context variables."""

    _elements = set()

    def __init__(self, __mapping=None, **kwargs):
        if __mapping is None:
            self.__values = dict(**kwargs)
        else:
            if isinstance(__mapping, BaseContext):
                __mapping = __mapping.__values
            self.__values = dict(__mapping)
            self.__values.update(**kwargs)

        bad_keys = set(self.__values) - self._elements

        if bad_keys:
            raise TypeError("Only %s keys are supported. %s given" %
                            (tuple(self._elements), tuple(bad_keys)))

    def __getattr__(self, name):
        try:
            return self.__values[name]
        except KeyError:
            if name in self._elements:
                return None
            else:
                raise AttributeError(name)

    def to_dict(self):
        return self.__values


class MistralContext(BaseContext):
    _elements = set([
        "user_id",
        "project_id",
        "auth_token",
        "service_catalog",
        "user_name",
        "project_name",
        "roles",
        "is_admin",
    ])


def has_ctx():
    return utils.has_thread_local(_CTX_THREAD_LOCAL_NAME)


def ctx():
    if not has_ctx():
        raise exc.ApplicationContextNotFoundException()

    return utils.get_thread_local(_CTX_THREAD_LOCAL_NAME)


def set_ctx(new_ctx):
    utils.set_thread_local(_CTX_THREAD_LOCAL_NAME, new_ctx)


def _wrapper(context, thread_desc, thread_group, func, *args, **kwargs):
    try:
        set_ctx(context)
        func(*args, **kwargs)
    except Exception as e:
        LOG.exception("Thread '%s' fails with exception: '%s'"
                      % (thread_desc, e))
        if thread_group and not thread_group.exc:
            thread_group.exc = e
            thread_group.failed_thread = thread_desc
    finally:
        if thread_group:
            thread_group._on_thread_exit()

        set_ctx(None)


def spawn(thread_description, func, *args, **kwargs):
    eventlet.spawn(_wrapper, ctx().clone(), thread_description,
                   None, func, *args, **kwargs)


def context_from_headers(headers):
    return MistralContext(
        user_id=headers.get('X-User-Id'),
        project_id=headers.get('X-Project-Id'),
        auth_token=headers.get('X-Auth-Token'),
        service_catalog=headers.get('X-Service-Catalog'),
        user_name=headers.get('X-User-Name'),
        project_name=headers.get('X-Project-Name'),
        roles=headers.get('X-Roles', "").split(",")
    )


class JsonPayloadSerializer(messaging.NoOpSerializer):
    @staticmethod
    def serialize_entity(context, entity):
        return jsonutils.to_primitive(entity, convert_instances=True)


class RpcContextSerializer(messaging.Serializer):

    def __init__(self, base=None):
        self._base = base or messaging.NoOpSerializer()

    def serialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.serialize_entity(context, entity)

    def deserialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.deserialize_entity(context, entity)

    def serialize_context(self, context):
        return context.to_dict()

    def deserialize_context(self, context):
        ctx = MistralContext(**context)
        set_ctx(ctx)
        return ctx


class ContextHook(hooks.PecanHook):
    def before(self, state):
        request_ctx = context_from_headers(state.request.headers)
        set_ctx(request_ctx)

    def after(self, state):
        set_ctx(None)



# -*- coding: utf-8 -*-
#
# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import functools

import pecan
import six
from wsme import exc

from mistral import exceptions as ex


def wrap_wsme_controller_exception(func):
    """This decorator wraps controllers method to manage wsme exceptions:
    In case of expected error it aborts the request with specific status code.
    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ex.MistralException as excp:
            pecan.response.translatable_error = excp
            raise exc.ClientSideError(msg=six.text_type(excp),
                                      status_code=excp.http_code)
    return wrapped


def wrap_pecan_controller_exception(func):
    """This decorator wraps controllers method to manage pecan exceptions:
    In case of expected error it aborts the request with specific status code.
    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ex.MistralException as excp:
            pecan.response.translatable_error = excp
            pecan.abort(excp.http_code, six.text_type(excp))
    return wrapped
